from etl.monitor import logger
from psycopg2.extras import execute_values
import json


def insert(db, schema, table, attrs, values):
    """
    Inserts a row into a specific table of the PG database.

    Parameters
    ----------
    db : obj
       : database connection object
    schema : string
           : name of schema
    table : string
          : name of table
    attrs : string[]
          : attributes (column names)
    values : string[]
           : values to insert; its length must be equal to len(attrs)
    Returns
    -------
    -

    Example
    -------
    insert('pg', 'public', 'company', attributes, values)
    TODO: schema should default to public
    """
    temp = []
    for v in values:
        if type(v) is list:
            if type(v[0]) is str and v[0].startswith("{"):
                temp.append('array[%s]::jsonb[]')
                continue
            temp.append('%s')

        else:
            temp.append('%s')

    temp = ', '.join(temp)
    attrs = ', '.join(attrs)

    cmd = "INSERT INTO %s.%s (%s) VALUES (%s) ON CONFLICT DO NOTHING;" % (
        schema, table.lower(), attrs, temp)

    # MoSQL ignores the document and logs a warning
    # if a document could not be inserted.
    # We will decide later what to do with DataErrors.
    db.execute_cmd(cmd, values)


def insert_bulk(db, schema, table, attrs, values):
    """
    Inserts multiple rows into a specific table of the PG database.

    Parameters
    ----------
    db : obj
       : database connection object
    schema : string
           : name of schema
    table : string
          : name of table
    attrs : string[]
          : attributes (column names)
    values : string[][]
           : values to insert
    Returns
    -------
    -

    Example
    -------
    insert('pg', 'public', 'company', attributes, values)
    TODO: schema should default to public
    """
    # prepare attributes for insert
    attrs = ', '.join([('"%s"' % a) for a in attrs])

    # default primary key in Postgres is name_of_table_pkey
    cmd = "INSERT INTO %s.%s (%s) VALUES %s;" % (
        schema, table.lower(), attrs, '%s')
    try:
        cur = db.conn.cursor()
        execute_values(cur, cmd, values)
        cur.close()
    except Exception as ex:
        logger.error("[ROW] INSERT bulk failed: %s" % ex)
        logger.error("[ROW] CMD: %s" % cmd)
        logger.error("[ROW] VALUES: %s" % values)
        raise SystemExit()


def upsert_bulk(db, schema, table, attrs, rows):
    """
    Inserts a row defined by attributes and values into a specific
    table of the PG database.

    Parameters
    ----------
    db : obj
    schema : string
    table_name : string
    attrs :     string[]
    rows :    string[]

    Returns
    -------
    -

    Example
    -------
    upsert_bulk(db, 'public', 'employee', [attributes], [values])
    Note: command is different for Postgres v10+:
    cmd = "INSERT INTO %s.%s (%s) VALUES (%s) ON CONFLICT ON CONSTRAINT %s
    DO UPDATE SET (%s) = ROW(%s);"
    """
    NR_OF_ROWS_TO_DISPLAY = 20
    temp = []
    for v in rows[0]:
        temp.append('%s')

    temp = ', '.join(temp)

    attrs_reduced = [('"%s"' % a) for a in attrs]
    attrs_reduced = ', '.join(attrs_reduced)

    excluded = [('EXCLUDED.%s' % a) for a in attrs]
    excluded = ', '.join(excluded)

    attrs = [('"%s"' % a) for a in attrs]
    attrs = ', '.join(attrs)

    # default primary key in Postgres is name_of_table_pkey
    constraint = '%s_pkey' % table

    cmd = """
    INSERT INTO %s.%s (%s) VALUES (%s) ON CONFLICT ON CONSTRAINT %s
    DO UPDATE SET (%s) = (%s);
    """ % (
        schema, table.lower(), attrs, temp,
        constraint, attrs_reduced, excluded)
    db.execute_many_cmd(cmd, rows)


def upsert_bulk_tail(db, schema, table, attrs, rows):
    """
    Inserts a row defined by attributes and values into a specific
    table of the PG database.
    Checks for values that are unset during tailing.

    Parameters
    ----------
    db : object
    schema : string
    table : string
    attrs : string[]
    rows : string[]

    Returns
    -------
    -

    Example
    -------
    upsert_bulk_tail(pg, 'public', 'employee', [attributes], [values])
    """
    for i in range(0, len(rows)):
        row = rows[i]
        values = []
        attrs_reduced = []
        for j in range(0, len(attrs)):
            if row[j] == '$unset':
                values.append(None)
                attrs_reduced.append(attrs[j])
            elif row[j] is not None:
                values.append(row[j])
                attrs_reduced.append(attrs[j])
        upsert_bulk(db, schema, table, attrs_reduced, [tuple(values)])




def upsert_transfer_info(db, schema, table, attrs, row):
    """
    Updates the collection map.
    """
    temp = []
    temp_row = []
    for r in row:
        if type(r) is list and type(r[0]) is dict:
            temp.append('%s::jsonb[]')
        else:
            temp.append('%s')

    placeholder = []

    for r in row:
        if type(r) is list and type(r[0]) is dict:
            for item in r:
                temp_row.append(json.dumps(item))
            placeholder.append("%s::jsonb[]")
        else:
            placeholder.append("'%s'" % r)

    placeholder = ','.join(placeholder)

    temp = ', '.join(temp)

    attrs_reduced = [('"%s"' % a) for a in attrs]
    attrs_reduced = ', '.join(attrs_reduced)

    excluded = [('EXCLUDED.%s' % a) for a in attrs]
    excluded = ', '.join(excluded)

    attrs = [('"%s"' % a) for a in attrs]
    attrs = ', '.join(attrs)

    # default primary key in Postgres is name_of_table_pkey
    constraint = '%s_pkey' % table

    # upsert
    cmd = """INSERT INTO %s.%s (%s) VALUES (%s) ON CONFLICT ON CONSTRAINT %s
    DO UPDATE SET (%s) = (%s);""" % (
        schema, table.lower(), attrs, placeholder,
        constraint, attrs_reduced, excluded)

    try:
        cur = db.conn.cursor()
        cur.execute(cmd, [temp_row])
    except Exception as ex:
        logger.error("[ROW] UPSERT TRANSFER INFO failed: %s" % ex)
        logger.error("\n[ROW] CMD:\n %s" % cmd)
        logger.error("\n[ROW] VALUES:\n %s" % row)
        raise SystemExit()
    cur.close()


def delete(db, schema, table_name, ids):
    """
    Deletes a row in a specific table of the PG database.

    Parameters
    ----------
    table_name : string
    object_id : ObjectId
                (will need to get the hex encoded version
                of ObjectId with str(object_id))

    Returns
    -------
    -

    Example
    -------
    delete(db, 'public', 'employee', "5acf593eed101e0c1266e32b")

    """
    oids = "','".join(ids)
    cmd = "DELETE FROM %s.%s WHERE id IN ('%s');" % (
        schema, table_name.lower(), oids)
    logger.info("[ROW] %s" % cmd)
    db.execute_cmd(cmd)



from etl.extract import collection
import unittest
from tests.meta import mock

mongo = mock.mongo
colls = ["Company", "Employee"]


def reset_dataset_mdb():
    # TODO
    print("resets collection company and employee")
    mongo.drop_collection(mock.coll_name_company)
    mongo.drop_collection(mock.coll_name_employee)
    docs = mock.data_mdb_company
    mongo[mock.coll_name_company].insert_many(docs)
    docs = mock.data_mdb_employee
    mongo[mock.coll_name_employee].insert_many(docs)


class TestCollections(unittest.TestCase):
    def test_check(self):
        reset_dataset_mdb()
        colls_in_db = collection.check(mongo, colls)
        print(colls)
        print(colls_in_db)
        assert colls == colls_in_db

    def test_get_by_name(self):
        reset_dataset_mdb()
        coll_name = "Company"
        docs = collection.get_by_name(mongo, coll_name)
        result = []
        for doc in docs:
            result.append({
                "active": doc["active"],
                "signupCode": doc["signupCode"],
                "domains": doc["domains"]
            })
        mocked = []
        for doc in mock.data_mdb_company:
            mocked.append({
                "active": doc["active"],
                "signupCode": doc["signupCode"],
                "domains": doc["domains"]
            })
        assert result == mocked

    def test_get_docs_for_type_check(self):
        reset_dataset_mdb()
        coll_name = "Employee"
        nr_of_docs = 2
        docs = collection.get_docs_for_type_check(
            mongo, coll_name, nr_of_docs)
        result = []
        for doc in docs:
            result.append(doc)
        length_equal = len(result) == nr_of_docs
        order_desc = result[0]['lastName'] == 'Hound'
        if length_equal is False:
            print("Number of requested documents differs: %d instead of %d" %
                  (len(result), nr_of_docs))
            assert False
        if order_desc is False:
            print("Order is not descending: %s", result)
            assert False
        assert True

    def test_get_by_name_reduced(self):
        reset_dataset_mdb()
        coll_name = "Employee"
        nr_of_docs = 2
        fields = ["firstName", "hair"]
        result = []
        docs = collection.get_by_name_reduced(
            mongo, coll_name, fields, nr_of_docs)

        for doc in docs:
            # leave out ObjectId
            result.append(
                {
                    "firstName": doc["firstName"],
                    "hair": doc["hair"]
                }
            )

        mocked = []
        for doc in mock.data_mdb_employee[::-1]:
            mocked.append(
                {
                    "firstName": doc["firstName"],
                    "hair": doc["hair"]
                }

            )
        print(result)
        print(mocked)
        if len(result) != len(mocked):
            assert False
        elif result != mocked:
            assert False
        else:
            return True

    def test_get_all(self):
        reset_dataset_mdb()
        mocked = mock.coll_names
        result = collection.get_all(mongo)
        assert mocked == result

    def test_get_doc_by_id(self):
        reset_dataset_mdb()
        coll_name = "Employee"
        obj = {"thing": 123}
        doc = mongo[coll_name].insert_one(obj)

        oid = str(doc.inserted_id)
        obj["_id"] = doc.inserted_id
        result = collection.get_doc_by_id(mongo, coll_name, oid)
        print(obj)
        print(result)
        assert result == obj



from bson.json_util import ObjectId
import datetime
import re


def get_type_pg(item):
    """
    Returns the item with its PG type.
    In case of list it detects if it consist of primitive types.
    If yes, it will return the corresponding type.
    Otherwise it will be a json array.
    Also, if a string is found, the ' will be replaced with '' so PG parser
    can know it's an apostrophe.
    """
    pg_type = None
    item_type = type(item)

    if item_type is bool:
        pg_type = 'boolean'

    elif item_type is int:
        pg_type = 'double precision'

    elif item_type is float:
        pg_type = 'double precision'

    elif item_type is str:
        pg_type = 'text'

    elif item_type is datetime.datetime:
        pg_type = 'timestamp'

    elif item_type is ObjectId:
        pg_type = 'text'

    elif item_type is dict:
        pg_type = 'jsonb'

    elif item_type is list:
        pg_type = 'jsonb'

    elif isinstance(item_type, type(None)):
        item = 'null'
    else:
        pg_type = 'jsonb'

    return item, pg_type


def type_equal(old, new):
    equal_char = ('char' in old and 'char' in new)
    equal_array = (old == 'array' and new == 'jsonb')
    equal_float = (old == 'double precision' and new == 'float')
    if equal_char or equal_array or equal_float:
        return True
    return False


def is_convertable(type_old, type_new):
    """
    Returns True if type old can be converted to type new

    convertables: list of tuples
                : contains convertable types (type_old, type_new)
    """
    convertables = [
        ('boolean', 'double precision'),
        ('boolean', 'jsonb'),
        ('boolean', 'timestamp'),
        ('boolean', 'text'),

        ('double precision', 'boolean'),  # TODO: test again
        # ('double precision', 'jsonb'), # cannot
        ('double precision', 'timestamp'),
        ('double precision', 'text'),

        ('jsonb', 'text'),  # TODO: test again
        ('jsonb', 'double precision'),
        ('jsonb', 'timestamp'),
        ('jsonb', 'text'),

        ('timestamp', 'boolean'),
        ('timestamp', 'double precision'),
        ('timestamp', 'jsonb'),
        ('timestamp', 'text'),  # works (tested)

        ('text', 'jsonb'),  # not possible
        ('text', 'boolean'),
        ('text', 'double precision'),
        ('text', 'timestamp'),  # works (tested)
    ]
    if (type_old.lower(), type_new.lower()) in convertables:
        return True
    return False


def snake_case(word):
    return re.sub('([A-Z]+)', r'_\1', word).lower().strip("_")



from etl.extract import collection
from etl.extract import tailer
from etl.extract import extractor
from etl.extract import transfer_info
import unittest
from tests.meta import mock
from bson import ObjectId
import datetime


mongo = mock.mongo
pg = mock.pg
query = mock.query
ex = extractor.Extractor(pg, mongo, mock.setup_pg,
                         mock.settings, mock.coll_config_company_employee)

settings_company = {
    ':columns':
    [
        {':source': '_id', ':type': 'TEXT', 'id': None},
        {':source': 'active', ':type': 'BOOLEAN', 'active': None},
        {':source': 'domains', ':type': 'JSONB', 'domains': None},
        {':source': 'signupCode', ':type': 'TEXT', 'signup_code': None}
    ],
    ':meta': {':extra_props': 'JSONB', ':table': 'company'}
}


def create_and_populate_company_pg():
    cursor = pg.conn.cursor()
    cursor.execute(query["table_drop_company"])
    cursor.execute(query["table_drop_employee"])
    query_company = """CREATE TABLE %s(%s, PRIMARY KEY (id));""" % (
        mock.rel_name_company, ', '.join(mock.attrs_types_company))

    cursor.execute(query_company)

    query_employee = """CREATE TABLE %s(%s, PRIMARY KEY (id));""" % (
        mock.rel_name_employee, ', '.join(mock.attrs_types_employee))

    cursor.execute(query_employee)
    cursor.close()


def reset_dataset():
    data_set_employee = [
        {
            "_id": "1",
            "firstName": "John",
            "lastName": "Snow",
            "hair": "black"
        },
        {
            "_id": "2",
            "firstName": "Arya",
            "lastName": "Start",
            "hair": "brown"
        }
    ]

    data_set_company = [
        {
            "_id": "1",
            "active": True,
            "signupCode": "uPsYdUpSy123",
            "domains": [
                "southpark.com"
            ]
        },
        {
            "_id": "2",
            "active": True,
            "signupCode": "node",
            "domains": [
                "amazon.com"
            ]
        }]

    mongo.drop_collection("Company")
    mongo.drop_collection("Employee")
    mongo["Company"].insert_many(data_set_company)
    mongo["Employee"].insert_many(data_set_employee)


class TestTailer(unittest.TestCase):
    def test_prepare_docs_for_update_one(self):
        docs = [{'coll_name': 'Company',
                 'db_name': 'test_purrito',
                 'o': {'$set': {'active': True, 'name': True}},
                 'o2': {'_id': ObjectId('5caf5998a54d758375bd9928')},
                 'op': 'u'}]
        mocked = [{'_id': '5caf5998a54d758375bd9928',
                   'active': True, 'name': True}]

        (docs_useful, merged) = tailer.prepare_docs_for_update(
            settings_company, docs)
        assert mocked == docs_useful and merged is False

    def test_prepare_docs_for_update_one_no_set_no_unset(self):
        docs = [{'coll_name': 'Company',
                 'db_name': 'test_purrito',
                 'o': {
                     '_id': ObjectId('5caf5998a54d758375bd9928'),
                     'active': True,
                     'name': True
                 },
                 'o2': {'_id': ObjectId('5caf5998a54d758375bd9928')},
                 'op': 'u'}]
        mocked = [{'_id': '5caf5998a54d758375bd9928',
                   'active': True, 'name': True, 'domains': '$unset', 'signupCode': '$unset'}]

        (docs_useful, merged) = tailer.prepare_docs_for_update(
            settings_company, docs)

        assert mocked == docs_useful and merged is False

    def test_prepare_docs_for_update_one_no_set_no_unset_all_nulls(self):
        docs = [{'coll_name': "Company",
                 'db_name': 'test_purrito',
                 'o': {
                     '_id': ObjectId('5caf5998a54d758375bd9928'),
                     'active': None
                 },
                 'o2': {'_id': ObjectId('5caf5998a54d758375bd9928')},
                 'op': 'u'}]
        mocked = [{
            '_id': '5caf5998a54d758375bd9928',
            'active': '$unset',
            'domains': '$unset',
            'signupCode': '$unset'
        }]
        (docs_useful, merged) = tailer.prepare_docs_for_update(
            settings_company, docs)
        assert mocked == docs_useful and merged is False

    def test_prepare_docs_for_update_multiple_merged_mixed(self):
        """
        This test covers the case when one document of a collection
        is updated directly through the IDE (e.g. Studio3T) and
        another is updated using an update query:
        coll.update({}, {"$set"}, {"$unset"})
        """
        docs = [{'coll_name': 'Company',
                 'db_name': 'test_purrito',
                 'o': {
                     '_id': ObjectId('5caf5998a54d758375bd9928'),
                     'active': True,
                     'name': True
                 },
                 'o2': {'_id': ObjectId('5caf5998a54d758375bd9928')},
                 'op': 'u'},

                {'coll_name': 'Company',
                 'db_name': 'test_purrito',
                 'o': {'$set': {'name': 'stuffy'}, '$unset': {'active': True}},
                 'o2': {'_id': ObjectId('5caf5998a54d758375bd9929')},
                 'op': 'u'}]

        mocked = [{'_id': '5caf5998a54d758375bd9928',
                   'active': True, 'name': True, 'domains': '$unset', 'signupCode': '$unset'},
                  {'_id': '5caf5998a54d758375bd9929',
                   'active': '$unset', 'name': 'stuffy'}]

        (docs_useful, merged) = tailer.prepare_docs_for_update(
            settings_company, docs)

        assert mocked == docs_useful and merged is False

    def test_prepare_docs_for_update_multiple_merged(self):
        docs = [{'coll_name': 'Company',
                 'db_name': 'test_purrito',
                 'o': {'$set': {'active': True, 'name': True}},
                 'o2': {'_id': ObjectId('5caf5998a54d758375bd9929')},
                 'op': 'u'},
                {'coll_name': 'Company',
                 'db_name': 'test_purrito',
                 'o': {'$set': {'name': 'stuffy'}, '$unset': {'active': True}},
                 'o2': {'_id': ObjectId('5caf5998a54d758375bd9929')},
                 'op': 'u'}]
        mocked = [{'_id': '5caf5998a54d758375bd9929',
                   'active': '$unset', 'name': 'stuffy'}]

        collection_settings = mock.coll_config
        collection_name = "Company"

        (docs_useful, merged) = tailer.prepare_docs_for_update(
            collection_settings[collection_name], docs)

        assert mocked == docs_useful and merged is True

    def test_handle_multiple(self):
        transfer_info.create_oplog_table(pg, 'public')
        reset_dataset()
        oplog_entries = mock.oplog_entries_update
        create_and_populate_company_pg()

        colls = ["Company", "Employee"]

        ex.transfer(colls)

        cursor = pg.conn.cursor()
        cursor.execute("SELECT * FROM public.employee where id = '1';")
        res = cursor.fetchall()

        t = tailer.Tailer(pg,
                          mongo,
                          mock.setup_pg,
                          mock.settings,
                          mock.coll_config_company_employee)
        ts = datetime.datetime.utcnow()-datetime.timedelta(minutes=2)

        t.handle_multiple(oplog_entries, ts)

        cursor = pg.conn.cursor()
        cursor.execute(
            "SELECT id, active, signup_code, domains FROM public.company WHERE id = '2';")
        res = cursor.fetchone()

        mocked_company = ("2", True, "node", ["dragonglass.org"])

        if res != mocked_company:
            for i in range(len(mocked_company)):
                if mocked_company[i] != res[i]:
                    assert False

        cursor.execute(
            "SELECT id, first_name, last_name, hair FROM public.employee WHERE id='1';")
        res = cursor.fetchone()

        mocked_janos = ('1', "Janos", None, None)

        if res != mocked_janos:
            for i in range(len(mocked_janos)):
                if mocked_janos[i] != res[i]:
                    assert False

        cursor.execute(
            "SELECT id, first_name, last_name, hair FROM public.employee WHERE id='2';")
        res = cursor.fetchone()
        cursor.close()

        mocked_arja = ("2", "Arja", "Stark", "blonde")

        if res != mocked_arja:
            for i in range(len(mocked_arja)):
                if mocked_arja[i] != res[i]:
                    assert False

        assert True

    def test_log_tailed_docs_one(self):
        transfer_info.create_oplog_table(pg)

        ids_log = ['58a32cfda51183070034909b']
        docs = [{
            '_id': '58a32cfda51183070034909b',
            'updatedAt': datetime.datetime(2019, 4, 29, 11, 26, 50, 703000)
        }]
        table_name = 'some_relation'
        oper = 'u'
        merged = True
        cursor = pg.conn.cursor()
        schema = 'public'
        tailer.log_tailed_docs(pg, schema, docs,
                               ids_log, table_name, oper,
                               merged)
        cursor.execute(
            """SELECT operation,
            relation, obj_id, merged, document FROM purr_oplog;""")
        res = cursor.fetchone()
        cursor.close()
        mock = tuple(
            ['u',
             'some_relation',
             '58a32cfda51183070034909b',
             True,
             "{'_id': '58a32cfda51183070034909b', 'updatedAt': datetime.datetime(2019, 4, 29, 11, 26, 50, 703000)}"
             ])
        assert mock == res

    def test_log_tailed_docs_multiple(self):
        transfer_info.create_oplog_table(pg)

        ids_log = ['58a32cfda51183070034909b', '58a32cfda51183070034909c']
        docs = [{
            '_id': '58a32cfda51183070034909b',
            'updatedAt': datetime.datetime(2019, 4, 29, 11, 26, 50, 703000)
        }, {
            '_id': '58a32cfda51183070034909c',
            'updatedAt': datetime.datetime(2019, 5, 21, 2, 24, 24, 5)
        }]
        table_name = 'some_relation'
        oper = 'u'
        merged = False
        cursor = pg.conn.cursor()
        schema = 'public'
        tailer.log_tailed_docs(pg, schema, docs,
                               ids_log, table_name, oper,
                               merged)
        cursor.execute(
            """SELECT operation,
            relation, obj_id, merged, document FROM purr_oplog order by id;""")
        res = cursor.fetchall()
        mock = [
            tuple(
                ['u',
                 'some_relation',
                 '58a32cfda51183070034909b',
                 False,
                 "{'_id': '58a32cfda51183070034909b', 'updatedAt': datetime.datetime(2019, 4, 29, 11, 26, 50, 703000)}"
                 ]),
            tuple(
                ['u',
                 'some_relation',
                 '58a32cfda51183070034909c',
                 False,
                 "{'_id': '58a32cfda51183070034909c', 'updatedAt': datetime.datetime(2019, 5, 21, 2, 24, 24, 5)}"
                 ])]
        cursor.close()
        assert mock == res



#!/usr/bin/env python3

import sys


def main():
    try:
        from etl.cli import main
        sys.exit(main())
    except KeyboardInterrupt:
        print('\nInterrupted by user.')


if __name__ == '__main__':
    main()



from etl.monitor import logger

CURR_FILE = "[PROCEDURE]"


def create_type_notification(db, name):
    """
    Creates a notification with a payload and sends it to channel 'purr'.

    Parameters
    ----------
    Returns
    -------
    Example
    -------
    create_notification(db)
    """
    cmd = """CREATE OR REPLACE FUNCTION %s()
    RETURNS TRIGGER AS $$
    BEGIN
        PERFORM pg_notify('purr', 'type_change');
        RETURN NULL;
    END;
    $$ LANGUAGE plpgsql;
    """ % name

    try:
        logger.info("Creating procedure: %s" % name, CURR_FILE)
        db.execute_cmd(cmd)

    except Exception as ex:
        logger.error("Insert failed: %s" % ex, CURR_FILE)


def drop_type_notification(db, name):
    """
    Drops a notification with a payload and sends it to channel 'purr'.

    Parameters
    ----------
    Returns
    -------
    Example
    -------
    drop_notification(db)
    """
    cmd = "DROP FUNCTION IF EXISTS %s();" % name

    try:
        db.execute_cmd(cmd)
        logger.info("Dropping procedure: %s" % name, CURR_FILE)
    except Exception as ex:
        logger.error("Dropping procedure failed: %s" % ex, CURR_FILE)



import unittest
from bson import ObjectId
from tests.meta import mock
from etl.extract import collection_map as cm

mongo = mock.mongo
colls = ["Company", "Employee"]


def reset_dataset_mdb():
    print("resets collection company and employee")
    mongo.drop_collection(mock.coll_name_company)
    mongo.drop_collection(mock.coll_name_employee)
    docs = mock.data_mdb_company
    mongo[mock.coll_name_company].insert_many(docs)
    docs = mock.data_mdb_employee
    mongo[mock.coll_name_employee].insert_many(docs)


class TestCollections(unittest.TestCase):
    def test_get_types(self):
        reset_dataset_mdb()
        docs = [
            {'_id': ObjectId('5caf1062a54d75f29d118eba'),
             'firstName': 'The',
             'lastName': 'Hound',
             'hair': 'brown'},
            {'_id': ObjectId('5caf1062a54d75f29d118eb9'),
             'firstName': 'Little',
             'lastName': 'Finger',
             'hair': 'dark brown'},
            {'_id': ObjectId('5caf1062a54d75f29d118eb8'),
             'firstName': 'Sansa',
             'lastName': 'Stark',
             'hair': 'ginger'},
            {'_id': ObjectId('5caf1062a54d75f29d118eb7'),
             'firstName': 'Arya',
             'lastName': 'Start',
             'hair': 'brown'},
            {'_id': ObjectId('5caf1062a54d75f29d118eb6'),
             'firstName': 'John',
             'lastName': 'Snow',
             'hair': 'black'}
        ]
        types = cm.get_types(docs)

        mocked_types = {
            '_id': {'text': 5},
            'firstName': {'text': 5},
            'lastName': {'text': 5},
            'hair': {'text': 5}
        }

        assert mocked_types == types



# access MongoDB
import pymongo
from etl.monitor import logger


class MongoConnection():
    """
    TODO
    - create a base class for Connection
    - put dbname somewhere else
    """

    def __init__(self, settings):
        db_name = settings['db_name']
        try:
            self.client = pymongo.MongoClient(settings['connection'])
        except Exception as ex:
            logger.error("Could not initialize MongoDB client: %s" % ex)
        try:
            self.conn = self.client[db_name]
        except Exception as ex:
            logger.error("Could not create connection to MongoDB: %s" % ex)

    def disconnect(self):
        logger.info("MongoDB says goodbye.")
        self.client.close()



import yaml
import etl.monitor as monitor


def config_basic(path):
    """
    Read a YAML file which contains the basic settings:
    Settings for postgres:
      schema_name: name of schema where the collections will be transfered
      schema_reset: drop and create the schema
      table_truncate: truncate tables before transfer
      table_drop: drop tables before transfer
      connection: connection string
        (example: postgres://127.0.0.1:5432/postgres)
    Settings for mongo:
      db_name: name of the database we want to connect to
      connection: connection string (example: mongodb://localhost:27017)
    tailing: if True, it starts tailing the oplog
    typecheck_auto: let Purr determine the type of each attribute
    without defining a config file (much slower!)
    Note
    ----
    If you decide to go without a setup file, most of the mentioned values
    will default to the most frequent user-choices so you can type less :)
    Default values:
    schema name: public
    schema_reset: false
    table_truncate: false
    table_drop: false
    tailing: false
    typecheck_auto: false

    Returns
    -------
    conf_file : dict
              : database name, user, schema_name, ...
    Parameters
    ----------
    path : string
         : path to file
    Example
    -------
    config_basic('path/to/file')
    """
    try:
        with open(path, 'r') as stream:
            conf_file = yaml.load(stream)
            if not conf_file:
                return None
            return conf_file
    except Exception as ex:
        monitor.logging.error(
            "[CONFIG PARSER] Failed to open setup file: %s" % ex)
        raise SystemExit()


def config_collections(path):
    """
      Read the collection map. The collection map is a YAML file which contains
      the user-defined settings for each relation:
      - name of MongoDB database
      - names of MongoDB collections and its field names
      - corresponding relation and attribute names/types in PG
      - types of extra properties
        - anything that is not defined by the user goes into _extra_props if it
        can be casted to the type _extra_props has

      Returns
      -------
      colls : dict
            : collection details
      Parameters
      ----------
      path : string
           : path to file
      Example
      -------
      config_collections('path/to/file')
      TODO
      ----
      Transfer data from multiple databases.
    """
    try:
        with open(path, 'r') as stream:
            conf_file = yaml.load(stream)
            db_names = list(conf_file.keys())
            colls = conf_file[db_names[0]]
            return colls
    except Exception as ex:
        monitor.logging.error(
            "[CONFIG PARSER] Failed to open collection file: %s" % ex)
        raise SystemExit()


def config_collection_names(colls):
    '''
    Get collection names from collection map (YAML file).
    Parameters
    ----------
    colls : dict
          : collections with all the field descriptions
    Returns
    -------
    coll_names : list
               : names of collections which should be transferred
    '''
    coll_names = list(colls.keys())
    if len(coll_names) == 0:
        monitor.logging.error(
            "[CONFIG PARSER] No collections found in the collection map.")
        raise SystemExit()
    return coll_names


def config_fields(colls, name):
    '''
    Get details based on collection.yml in order to prepare rows for pg.
    Returns
    -------
    attrs_new: names of columns in PG
    attrs_original: names of fields from MongoDB
    types : column types
    relation_name: names of relation in PG
    extra_props_type: type of extra properties (preferred JSONB)

    Parameters
    ----------
    colls : dict
          : collection names with settings from collections.yml
    name : string
         : name of collection
    Example
    -------
    config_fields(collections_with_settings, collection_name)
    '''
    attrs_original = []
    attrs_new = []
    types = []

    if name in colls.keys():
        collection = colls[name]
    else:
        monitor.logging.warn(
            """
            [CONFIG PARSER] Failed to find description of %s in
            collections.yml. Skipping collection.
            """ % name)
        return ([], [], [], [], [])
    relation_name = collection[":meta"][":table"]
    extra_props_type = collection[":meta"][":extra_props"]
    for field in collection[":columns"]:
        for key, value in field.items():
            if key.startswith(":") is False:
                attrs_new.append(key)
            else:
                if key == ":source":
                    attrs_original.append(value)
                elif key == ":type":
                    types.append(value)
    return attrs_new, attrs_original, types, relation_name, extra_props_type



from etl.monitor import logger


def reset(db, schema='public'):
    """
    Reset existing schema or create a new one.
    """
    drop = 'DROP SCHEMA IF EXISTS %s CASCADE;' % schema
    create = 'CREATE SCHEMA %s;' % schema
    try:
        db.execute_cmd(drop)
        db.execute_cmd(create)
        logger.info("[SCHEMA] Schema %s is reset." % schema)
    except Exception as ex:
        logger.error("[SCHEMA] Schema reset failed. %s" % ex)


def create(db, schema='public'):
    """
    Create schema if it does not exist.
    """
    cmd = 'CREATE SCHEMA IF NOT EXISTS %s;' % (schema)
    try:
        db.execute_cmd(cmd)
    except Exception as ex:
        logger.error("""
        [SCHEMA] Creating schema with name %s failed.
        Details: %s
        """ % (schema, ex))



import logging
import sys


class Logger():
    """Class for logs."""

    def __init__(self, name=""):
        """
          Creates log handlers, adds formatters to them
          and finally adds the handlers to the logger.
        """
        self.logger = logging.getLogger()
        self.logger.setLevel(logging.DEBUG)
        formatter = logging.Formatter(
            "%(asctime)s - [%(levelname)s] - %(message)s")

        h = logging.StreamHandler(sys.stdout)
        h.setLevel(logging.DEBUG)
        h.setFormatter(formatter)

        self.logger.addHandler(h)

    def info(self, m, location=""):
        self.logger.info(location + " " + m)

    def warn(self, m, location=""):
        self.logger.warning(location + " " + m)

    def debug(self, m, location=""):
        self.logger.debug(location + " " + m)

    def error(self, m, location=""):
        self.logger.error(location + " " + m)

    def critical(self, m, location=""):
        self.logger.critical(location + " " + m)


logger = Logger()



from etl.monitor import logger
import pymongo
from bson import ObjectId


def check(db, colls_requested):
    """
    Checks if requested collections exist in the database.
    Gets all collection names from MongoDB (colls_name) and creates
    a new list which contains only the existing collection names.

    Parameters
    ----------
    db : pymongo.database.Database
      Database connection and name
    colls_requested : list
      Contains the list of requested collection names.

    Returns
    -------
    colls_existing : list
      Contains only existing collection names.

    Raises
    ------
    ValueError
      when a requested collection does not exist in the database (MongoDB)

    Example
    -------
    check(db, ['Car'])
      []
    check(db, ['Region', 'Customer']
      ['Region', 'Customer']
    """
    colls_name = db.collection_names(include_system_collections=False)
    colls_existing = []
    logger.info('[COLLECTION] Checking collection names...')
    try:
        for coll in colls_requested:
            try:
                colls_name.index(coll)
                colls_existing.append(coll)
            except ValueError:
                logger.warn("""
                    [COLLECTION] '%s' is not in the Mongo database. 
                    Skipping data transfer""" % coll)
    except Exception as ex:
        logger.error("[COLLECTION] Checking collection names failed: %s" % ex)
    return colls_existing


def get_by_name(db, name, size=20000):
    """
    Gets data from collection limited by batch size.

    Parameters
    ----------
    db : pymongo.database.Database
      Database connection and name
    name : string
      Name of collection.

    Returns
    -------
    docs : pymongo.cursor.Cursor

    Raises
    ------

    Example
    -------
    get_by_name(db, 'Car')

    TODO
    ----
    - let the user decide batch size
    """
    docs = []
    try:
        logger.info('[COLLECTION] Loading data from collection %s...' % name)
        c = db[name]
        bz = c.find()
        docs = bz.batch_size(size)
    except Exception as ex:
        logger.error(
            """[COLLECTION] Loading data from collection %s failed.
            Details: %s""" % (name, ex))
    return docs


def get_docs_for_type_check(db, name, nr_of_docs=100):
    """
    Gets data from a collection limited.

    Parameters
    ----------
    db : pymongo.database.Database
      Database connection
    name : string
      Name of collection.
    nr_of_docs : integer
      Number of documents to return

    Returns
    -------
    docs : pymongo.cursor.Cursor

    Raises
    ------

    Example
    -------
    get_docs_for_type_check(db, 'Car')

    TODO
    ----
    - let the user decide batch size
    """
    docs = []
    try:
        logger.info('[COLLECTION] Loading data from collection %s...' % name)
        c = db[name]
        docs = c.find().sort(
            '$natural', pymongo.DESCENDING
        ).skip(0).limit(nr_of_docs)
    except Exception as ex:
        logger.error("""
          [COLLECTION] Loading data from collection %s failed.
          Details: %s""" % (name, ex))
    return docs


def get_by_name_reduced(db, name, fields, size=20000):
    """
    Gets data from collection limited by batch size containing
    only specific fields.

    Parameters
    ----------
    db : pymongo.database.Database
      Database connection and name
    name : string
      Name of collection.
    fields : list
      Names of fields to include in the query.

    Returns
    -------
    docs : pymongo.cursor.Cursor

    Raises
    ------

    Example
    -------
    get_by_name(db, 'Car', ['_id', 'type', 'nfOfSeats'])

    TODO
    ----
    - let the user decide batch size
    """
    docs = []
    try:
        logger.info('[COLLECTION] Loading data from collection %s...' % name)
        c = db[name]
        # create the document given to a query that specifies which
        # fields MongoDB returns in the result set
        projection = {}
        for field in fields:
            projection[field] = 1
        bz = c.find({}, projection).sort('$natural', pymongo.DESCENDING)
        docs = bz.batch_size(size)
    except Exception as ex:
        logger.error(
            """[COLLECTION] Loading data from collection %s failed.
            Details: %s""" % (name, ex))
    return docs


def get_all(db):
    try:
        return db.collection_names(include_system_collections=False)
    except Exception as ex:
        logger.error(
            """[COLLECTION] Loading collection names failed.
            Details:%s""" % (ex))


def get_doc_by_id(db, name, id):
    try:
        c = db[name]
        bz = c.find_one({"_id": ObjectId(id)})
        return bz
    except Exception as ex:
        logger.error(
            """
            [COLLECTION] Loading document from collection %s failed.
            Details: %s""" % (name, ex))



# here comes everything with the oplog
import pymongo
import time
from datetime import datetime
from bson import Timestamp

from etl.transform import relation
from etl.monitor import logger
from etl.extract import extractor, transfer_info
import json

INSERT = "i"
UPDATE = "u"
DELETE = "d"

CURR_FILE = "[TAILER]"


def prepare_docs_for_update(coll_settings, docs):
    docs_useful = []
    docs_id = []
    for doc in docs:
        # It is possible that multiple versions of one document
        # exist among these documents. they must be merged so they
        # can be sent Postgres together as one entry.
        merge_similar = False
        unset = {}
        doc_useful = {}
        temp = doc["o"]

        if "o2" in doc.keys():
            if "_id" in doc["o2"].keys():
                doc_useful["_id"] = str(doc["o2"]["_id"])
                if (doc_useful["_id"] in docs_id):
                    merge_similar = True
                else:
                    docs_id.append(str(doc_useful["_id"]))

        if "$set" in temp.keys():
            doc_useful.update(temp["$set"])
            for k, v in temp["$set"].items():
                if v is None:
                    unset[k] = "$unset"
        if "$unset" in temp.keys():
            for k, v in temp["$unset"].items():
                unset[k] = '$unset'
        if "$set" not in temp.keys() and "$unset" not in temp.keys():
            # case when the document was not updated
            # using a query, but the IDE e.g. Studio3T:
            logger.info("Direct update:")
            doc_useful.update(temp)
            fields = [x[":source"]
                      for x in coll_settings[":columns"]]
            for k in fields:
                if k == '_id':
                    temp[k] = str(temp[k])
                    doc_useful.update(temp)
                if k not in temp.keys():
                    unset[k] = '$unset'
            for k, v in temp.items():
                if v is None:
                    unset[k] = '$unset'
        doc_useful.update(unset)

        # merging values with the same ID because there cannot be
        # multiple updates of the same row in one statement
        if merge_similar is True:
            for i in range(0, len(docs_useful)):
                if docs_useful[i]["_id"] == doc_useful["_id"]:
                    docs_useful[i] = dict(docs_useful[i], **doc_useful)
                    break
        else:
            docs_useful.append(doc_useful)
    return docs_useful, merge_similar


def log_tailed_docs(pg, schema, docs_useful, ids_log, table_name, oper, merged):
    log_entries = []
    ts = time.time()
    logger.info("IDs: %s" % ids_log)
    if len(ids_log) != len(docs_useful) and oper != 'd':
        logger.error("n(ids)=%s; n(docs_useful)=%s" %
                     (len(ids_log), len(docs_useful)))
    for i in range(len(docs_useful)):
        id = ids_log[i]
        doc = "no entry"
        try:
            if docs_useful[i] is not None and oper != 'd':
                doc = str(docs_useful[i])
            else:
                doc = "Doc is NULL"
        except Exception as ex:
            logger.error("%s Converting log entry failed. Details: %s\n Document: " %
                         (CURR_FILE, ex))
            logger.error(docs_useful[i])
        row = [oper, table_name, id, ts,
               merged, doc]
        log_row = tuple(row)
        log_entries.append(log_row)
    try:
        transfer_info.log_rows(pg, schema, log_entries)
    except Exception as ex:
        logger.error("%s Logging failed. Details: %s" % (CURR_FILE, ex))


class Tailer(extractor.Extractor):
    """
    Class for extracting data from the oplog.
    """

    def __init__(self, pg, mdb, setup_pg, settings, coll_settings):
        """
        Parameters
        ----------
        pg : postgres connection
        mdb : mongo connection
        setup_pg : postgres specific settings from setup.yml
        settings : settings from setup.yml
        coll_settings : settings for each collection from collections.yml
        """
        extractor.Extractor.__init__(
            self, pg, mdb, setup_pg, settings, coll_settings)
        self.pg = pg
        self.schema = setup_pg["schema_name"]
        self.settings = settings
        self.coll_settings = coll_settings
        self.stop_tailing = False

    def coll_in_map(self, name):
        '''
        name: string;
            name of collection as 'name_db.name_coll', e.g. 'cat_db.Breeds'
        Checks if a collection exists in collections.yml.
        '''
        coll = name.split(".")[1]
        try:
            if coll in self.coll_settings.keys():
                return True
            else:
                return False
        except Exception as ex:
            logger.error("%s Deatils %s" % (CURR_FILE, ex))
            return False

    def flush(self, docs, oper, r):
        """
        sends all the data which was collected for one
        collection during tailing to Postgres
        """
        docs_useful = []
        ids_log = []
        merged = False

        if oper == INSERT:
            logger.info("%s Inserting %s documents into '%s'" %
                        (CURR_FILE, len(docs), r.relation_name))

            # TODO: check extra props
            for doc in docs:
                ids_log.append(str(doc["o"]["_id"]))
                docs_useful.append(doc["o"])
            try:
                super().insert_multiple(docs_useful, r, docs[0]["coll_name"])
            except Exception as ex:
                logger.info(
                    """
                    %s Inserting multiple documents failed: %s.
                    Details: %s
                    """ % (CURR_FILE, docs, ex))

        elif oper == UPDATE:
            logger.info("%s Updating %s documents in '%s'" %
                        (CURR_FILE, len(docs), r.relation_name))
            r.created = True

            coll_name = docs[0]["coll_name"]
            coll_settings = self.coll_settings[coll_name]
            (docs_useful, merged) = prepare_docs_for_update(
                coll_settings,
                docs
            )
            for doc in docs_useful:
                ids_log.append(str(doc["_id"]))
            try:
                super().update_multiple(docs_useful, r, docs[0]["coll_name"])
            except Exception as ex:
                logger.info(
                    """%s Updating multiple documents failed: %s.
                    Details: %s""" % (CURR_FILE, docs, ex))

        elif oper == DELETE:
            logger.info("%s Deleting %s documents from '%s'" %
                        (CURR_FILE, len(docs), r.relation_name))
            ids = []
            for doc in docs:
                ids.append(doc["o"])
                ids_log.append(str(doc["o"]["_id"]))
            try:
                r.delete(ids)
            except Exception as ex:
                logger.info(
                    """%s Deleting multiple documents failed: %s.
                    Details: %s""" % (CURR_FILE, docs, ex))

        log_tailed_docs(
            self.pg, self.schema,
            docs_useful, ids_log,
            r.relation_name, oper, merged
        )

    def transform_and_load_many(self, docs_details):
        """
        Gets the document and passes it to the corresponding
        function in order to exeucte command INSERT/UPDATE/DELETE
        Parameters
        ----------
        docs :  list of documents and operations
                document from MongoDB
        Example
        -------
        transform_and_load({"_id":ObjectId("5acf593eed101e0c1266e32b"))
        Return
        ------
        None
        """
        table_name_mdb = docs_details[0]["coll_name"]
        table_name_pg = self.coll_settings[table_name_mdb][":meta"][":table"]

        r = relation.Relation(self.pg, self.schema, table_name_pg, True)

        oper = docs_details[0]["op"]

        docs_with_equal_oper = []
        for doc_details in docs_details:
            if oper == doc_details["op"]:
                docs_with_equal_oper.append(doc_details)
            else:
                self.flush(docs_with_equal_oper, oper, r)
                oper = doc_details["op"]
                docs_with_equal_oper = [doc_details]
        self.flush(docs_with_equal_oper, oper, r)

    def handle_multiple(self, docs, updated_at):
        # group by name
        docs_grouped = {}
        for doc in docs:
            if doc["ns"] not in docs_grouped.keys():
                docs_grouped[doc["ns"]] = []

            useful_info_update = {}
            if "o2" in doc.keys():
                useful_info_update = doc["o2"]

            docs_grouped[doc["ns"]].append({
                "op": doc["op"],
                "db_name": doc["ns"].split(".")[0],
                "coll_name": doc["ns"].split(".")[1],
                "o": doc["o"],
                "o2": useful_info_update
            })

        for coll, docs_details in docs_grouped.items():
            self.transform_and_load_many(docs_details)
            # every 5 minutes update the timestamp because we need to continue
            # tailing in case of disconnecting from the PGDB
            diff = datetime.utcnow() - updated_at
            minutes_between_update = (
                diff.seconds//60) % 60
            if minutes_between_update > 5:
                t = int(time.time())
                transfer_info.update_latest_successful_ts(
                    self.pg, self.schema, t
                )
                logger.info(
                    "%s Updated latest_successful_ts: %d" % (CURR_FILE, t))

    def start(self, dt=None):
        """
        Starts tailing the oplogs.
        Parameters
        ----------
        dt: datetime
            If datetime is None, tailing starts from the current timestamp.
        Example
        -------
        (1)
        start()

        (2)
        start(dt)
        """
        client = self.mdb.client
        oplog = client.local.oplog.rs

        # Start reading the oplog
        SECONDS_BETWEEN_FLUSHES = 30
        try:
            updated_at = datetime.utcnow()
            loop = False
            while True and self.stop_tailing is False:
                if loop is True:
                    res = transfer_info.get_latest_successful_ts(
                        self.pg, self.schema)
                    latest_ts = int((list(res)[0])[0])
                    dt = latest_ts
                    logger.error(
                        """%s Stopping. Next time, bring more cookies."""
                        % (CURR_FILE))
                    raise SystemExit
                else:
                    loop = True

                # if there was a reconnect attempt then start tailing from
                # specific timestamp from the db
                if self.pg.attempt_to_reconnect is True:
                    res = transfer_info.get_latest_successful_ts(
                        self.pg, self.schema)
                    dt = int((list(res)[0])[0])
                    self.pg.attempt_to_reconnect = False

                cursor = oplog.find(
                    {"ts": {"$gt": Timestamp(dt, 1)}},
                    cursor_type=pymongo.CursorType.TAILABLE_AWAIT,
                    oplog_replay=True,
                )
                if type(dt) is int:
                    dt = datetime.utcfromtimestamp(
                        dt).strftime('%Y-%m-%d %H:%M:%S')
                logger.info("%s Started tailing from %s.\nCurrent timestamp: %s" %
                            (CURR_FILE, str(dt), datetime.utcnow()))

                docs = []
                while cursor.alive and self.pg.attempt_to_reconnect is False:
                    if self.stop_tailing is True:
                        logger.info(
                            "%s Meow" % (CURR_FILE))
                        break
                    try:
                        for doc in cursor:
                            col = doc["ns"]
                            op = doc["op"]
                            if op != "n" and self.coll_in_map(col) is True:
                                docs.append(doc)
                        time.sleep(1)
                        seconds = datetime.utcnow().second
                        if ((seconds > SECONDS_BETWEEN_FLUSHES and len(docs)) or (len(docs) > 100)):
                            logger.info("""
                            %s Flushing after %s seconds.
                            Number of documents: %s
                            """ % (CURR_FILE, seconds, len(docs)))
                            self.handle_multiple(docs, updated_at)
                            docs = []
                    except Exception as ex:
                        logger.error("%s Cursor error: %s" % (CURR_FILE, ex))
                cursor.close()
                continue
        except StopIteration as e:
            logger.error("%s Tailing was stopped unexpectedly: %s" %
                         (CURR_FILE, e))

    def stop(self):
        logger.info("%s Tailing is stopped due to schema change." %
                    (CURR_FILE))
        self.stop_tailing = True



from etl.load import table, row, procedure
from etl.monitor import logger
from datetime import datetime
import collections
from etl.extract import collection
from etl.transform import type_checker as tc

import yaml

CURR_FILE = "[COLLECTION_MAP]"


def get_query_update(def_coll, schema, table, attrs, row_id):
    types = []
    for field in def_coll:
        tmp = (str(field).replace("'", '"').replace('None', 'null'))
        types.append("$$%s$$" % tmp)
    query = """
    UPDATE %s.%s SET types = ARRAY[%s]::JSONB[]
    WHERE id=%s;
    """ % (schema, table, ", ".join(types), row_id)
    return query


def populate_table(db, coll_map, table, attrs, schema='public'):
    # TODO: default to public
    collection_map = collections.OrderedDict(coll_map)
    for coll_name, v in coll_map.items():
        row_id = list(collection_map).index(coll_name)
        query_update = get_query_update(
            v[":columns"], schema, table, attrs, row_id
        )
        values = tuple([row_id,
                        coll_name,
                        v[":meta"][":table"],
                        v[":columns"],
                        datetime.utcnow(),
                        query_update])
        row.upsert_transfer_info(db, schema, table, attrs, values)


def get_table(db, schema='public'):
    cmd = """SELECT id, collection_name, relation_name,
    types FROM %s.purr_collection_map ORDER BY id""" % (
        schema)
    try:
        coll_map = db.execute_cmd_with_fetch(cmd)
        logger.info("Getting schema from DB.", CURR_FILE)
        return coll_map

    except Exception as ex:
        logger.error(
            "[TRANSFER_INFO] Failed to get collection map table"
            % (ex)
        )


def create_table(db, coll_map, schema='public'):
    """
  Adds primary key to a PostgreSQL table.
  Parameters
  ----------
  Returns
  -------
  -
  Example
  -------
  create_table(pg, 'purr')
  """
    table_name = "purr_collection_map"
    attrs = ["id", "collection_name", "relation_name",
             "types", "updated_at", "query_update"]
    types = ["integer", "text", "text", "jsonb[]", "timestamp", "text"]

    try:
        # TODO: make this fucntion accept string and list
        table.drop(db, schema, [table_name])
        table.create(db, schema, table_name, attrs, types)
        logger.info("Created table %s.%s." %
                    (schema, table_name), CURR_FILE)
    except Exception as ex:
        logger.error(
            """
            Failed to create table %s.%s: %s
            """ % (schema, table_name, ex), CURR_FILE)

    populate_table(db, coll_map, table_name, attrs, schema)
    procedure_name = 'notify_type'
    procedure.drop_type_notification(db, procedure_name)
    procedure.create_type_notification(db, procedure_name)
    table.drop_trigger_type_notification(
        db, 'public', 'purr_collection_map', 'notify', procedure_name)
    table.create_trigger_type_notification(
        db, 'public', 'purr_collection_map', 'notify', procedure_name)


def get_types(docs):
    types = {}
    for doc in docs:
        fields = []
        for k, v in doc.items():
            name_column = tc.snake_case(k)
            if name_column in fields:
                logger.warn(
                    "Column %s cannot appear twice. Skipping..."
                    % k, CURR_FILE)
                continue
            else:
                fields.append(name_column)

            if k not in types.keys():
                types[k] = {}

            value_new, type_pg = tc.get_type_pg(v)
            if type_pg in types[k].keys():
                types[k][type_pg] += 1
            else:
                types[k][type_pg] = 1
    return types


def create_map(mongo, name_db):
    coll_map = {name_db: {}}
    colls = collection.get_all(mongo)

    for coll in sorted(colls):
        logger.info(
            'Determining types for collection %s...' %
            coll, CURR_FILE)

        # TODO: replace snake_case to another file e.g util
        name_relation = tc.snake_case(coll)

        coll_map[name_db][coll] = {
            ":columns": [],
            ':meta': {
                ':table': name_relation,
                ':extra_props': 'JSONB'
            }
        }
        docs = collection.get_docs_for_type_check(mongo, coll)
        logger.info("Reading samples...", CURR_FILE)

        types = get_types(docs)

        # TODO: handle None
        for field, value in types.items():
            type_chosen = "text"
            if len(field) > 1:
                sum = docs.count()
                max = 0
                for k, v in value.items():
                    if k is None:
                        continue
                    curr_perc = v/sum
                    if curr_perc > 0 and curr_perc > max:
                        max = curr_perc
                        type_chosen = k
            else:
                # there is exactly one key which will be the
                # chosen type
                type_chosen = list(value.keys())[0]
            name_column = tc.snake_case(field)
            def_column = {
                name_column: None,
                ":source": field,
                ":type": type_chosen.upper()
            }
            coll_map[name_db][coll][":columns"].append(def_column)
    return coll_map


def create_file(coll_map):
    """
    Creates the collection map file.
    """
    name_file = "collections.yml"
    operation = "w"
    try:
        logger.info("%s Creating collection map file..." % CURR_FILE)
        with open(name_file, operation) as file_out:
            yaml.dump(coll_map, file_out, default_flow_style=False)
        logger.info(
            "Collection map file created: %s" % name_file, CURR_FILE)
    except Exception as ex:
        logger.error("Failed to create collection map file. Details: %s" %
                     ex, CURR_FILE)



from etl.extract import collection
from etl.extract import collection_map as cm
from etl.load import table, schema
from etl.monitor import logger
from etl.extract import transfer_info
import time
from etl.transform import relation, type_checker as tc, config_parser as cp

name_extra_props_pg = "_extra_props"
name_extra_props_mdb = "extraProps"

CURR_FILE = "[EXTRACTOR]"


class Extractor():
    """
    This is a class for extracting data from collections.
    """

    def __init__(self, pg, mdb, settings_pg, settings_general, coll_def):
        """Constructor for Extractor"""

        self.pg = pg
        self.mdb = mdb
        self.include_extra_props = settings_general['include_extra_props']
        try:
            self.include_extra_props = settings_general['include_extra_props']
        except KeyError:
            self.include_extra_props = False
        self.schema = settings_pg["schema_name"]
        self.truncate = settings_pg['table_truncate']
        self.drop = settings_pg['table_drop']
        self.coll_def = coll_def
        self.tailing_from = settings_general['tailing_from']
        self.tailing_from_db = settings_general['tailing_from_db']
        self.coll_map_cur = cm.get_table(self.pg)
        self.attr_details = {}

    def convert_columns(self, name_table, source, fields_cur, fields_new):
        """
        (1) Tries to convert the column
        (2) TODO: If (1) was not successful (PG could not
        convert the column), just rename it and add
        the column again so Purr can take care of it
        """
        for i in range(0, len(fields_new)):
            field = fields_new[i]
            if field[":source"] in source:
                for column, v in field.items():
                    if v is None:
                        type_old = fields_cur[i][":type"]
                        type_new = field[":type"]
                        if tc.is_convertable(type_old, type_new):
                            logger.info(
                                """%s table %s, column %s:
                                Type [%s] is convertable to [%s]""" % (
                                    CURR_FILE,
                                    name_table,
                                    column,
                                    type_old,
                                    type_new
                                ))
                            table.column_change_type(
                                self.pg,
                                self.schema,
                                name_table,
                                column,
                                type_new)
                        else:
                            logger.error("""
                                %s In table %s, column %s:
                                Type [%s] is NOT convertable to [%s]
                                """ % (
                                CURR_FILE,
                                name_table,
                                column,
                                type_old,
                                type_new))

    def add_columns(self, source, added, name_coll, name_table, fields_new, cm):
        """
        Adds columns to a table and updates the coll_def.
        After that, it restarts the collection transfer
        so the changes would be picked up.
        """
        for item in added:
            if item[":source"] not in source:
                for attribute, v in item.items():
                    if v is None:
                        self.coll_def[name_coll][":columns"] = fields_new
                        table.add_column(
                            self.pg,
                            self.schema,
                            name_table,
                            attribute,
                            item[":type"])
                        self.transfer_coll(cm[1])

    def remove_columns(self, source, removed, name_coll, name_table, fields_new):
        """
        Removes columns to a table and updates the coll_def.
        """
        for item in removed:
            if item[":source"] not in source:
                for attribute, v in item.items():
                    if v is None:
                        # update collection settings
                        self.coll_def[name_coll][":columns"] = fields_new
                        table.remove_column(
                            self.pg,
                            self.schema,
                            name_table,
                            attribute)

    def update_table_def(self, coll_map_cur, coll_map_new):
        """
        Updates tables' columns and collection definition (coll_def)
        """
        for i in range(0, len(coll_map_cur)):
            name_coll = coll_map_cur[i][1]
            name_table = coll_map_cur[i][2]
            fields_cur = coll_map_cur[i][3]
            fields_new = coll_map_new[i][3]

            removed = [x for x in fields_cur if x not in fields_new]
            added = [x for x in fields_new if x not in fields_cur]

            sources_removed = [x[":source"] for x in removed]
            sources_added = [x[":source"] for x in added]

            source_persistent = [
                x for x in sources_removed if x in sources_added]

            if len(source_persistent):
                self.convert_columns(
                    name_table,
                    source_persistent,
                    fields_cur,
                    fields_new
                )

            self.add_columns(source_persistent, added,
                             name_coll, name_table, fields_new,
                             coll_map_cur[i])

            self.remove_columns(source_persistent, removed,
                                name_coll, name_table, fields_new)

    def table_track(self, coll_map_cur, coll_map_new):
        """
        coll_map_cur : list
                     : current collection map
        coll_map_new : list
                     : new collection map
        Update the extractor object's collection map and
        starts tracking collections (data transfer).

        TODO: take care of extra props type (JSONB)
        """
        logger.info("%s Adding new collection" %
                    CURR_FILE)
        colls_cur = [x[1] for x in coll_map_cur]
        colls_new = [x[1] for x in coll_map_new]
        colls_to_add = [x for x in colls_new if x not in colls_cur]

        for name_coll in colls_to_add:
            coll_def = [x for x in coll_map_new if x[1] == name_coll][0]
            columns = coll_def[3]
            name_rel = coll_def[2]
            type_extra_prop = 'JSONB'

            meta = {
                ':table': name_rel,
                ':extra_props': type_extra_prop
            }

            self.coll_def[name_coll] = {
                ':columns': columns,
                ':meta': meta,
            }
            self.transfer_coll(name_coll)

    def table_untrack(self, coll_map_cur, coll_map_new):
        tables_cur = [x[2] for x in coll_map_cur]
        tables_remaining = [x[2] for x in coll_map_new]

        tables_to_drop = [x for x in tables_cur if x not in tables_remaining]
        colls_to_remove = [x[1]
                           for x in coll_map_cur if x[2] in tables_to_drop]

        logger.info("%s Stop syncing collections %s." %
                    (
                        CURR_FILE,
                        ", ".join(colls_to_remove)))

        for coll in colls_to_remove:
            self.coll_def.pop(coll, None)

    def update_coll_map(self):
        # TODO: get types from pg
        logger.info("%s Updating schema from purr_collection_map" %
                    CURR_FILE)

        coll_map_cur = self.coll_map_cur
        coll_map_new = cm.get_table(self.pg)

        if coll_map_cur == coll_map_new:
            logger.info("%s Database schema is not changed" %
                        CURR_FILE)
            return

        # If no tables were added or removed then update
        # table definition, otherwise add or remove a table.
        if len(coll_map_new) == len(coll_map_cur):
            self.update_table_def(coll_map_cur, coll_map_new)
        elif len(coll_map_new) > len(coll_map_cur):
            self.table_track(coll_map_cur, coll_map_new)
        else:
            self.table_untrack(coll_map_cur, coll_map_new)

        # update current collection map - from the db
        self.coll_map_cur = coll_map_new

        # TODO:
        # checking if the relation name is changed
        # restart transfer
        # update schema

    def transfer(self, coll_names_in_config):
        """
        Transfers documents or whole collections if the number of fields
        is less than 30 000 (batch_size).
        Types of attributes are determined using the collections.yml file.
        Returns
        -------
        -
        Parameters
        ----------
        coll_names : list
                   : list of collection names
        """

        coll_names = collection.check(self.mdb, coll_names_in_config)
        if len(coll_names) == 0:
            logger.info('%s No collections to transfer.' %
                        CURR_FILE
                        )
            return

        relation_names = []
        for coll in coll_names:
            relation_names.append(tc.snake_case(coll))

        if self.drop:
            table.drop(self.pg, self.schema, relation_names)
        elif self.truncate:
            table.truncate(self.pg, self.schema, coll_names)

        schema.create(self.pg, self.schema)

        for coll in coll_names:
            self.transfer_coll(coll)

    def transfer_coll(self, coll):
        '''
        Transfers documents or whole collections if the number of fields
        is less than 30 000 (batch_size).
        Returns
        -------
        -
        Parameters
        ----------
        coll : string
             : name of collection which is going to be transferred
        '''

        r = self.adjust_columns(coll)

        if self.tailing_from is not None or self.tailing_from_db is True:
            return

        if self.include_extra_props is True:
            docs = collection.get_by_name(self.mdb, coll)
        else:
            attr_source = [k for k, v in self.attr_details.items()]
            docs = collection.get_by_name_reduced(self.mdb, coll, attr_source)

        # Start transferring docs
        nr_of_docs = docs.count()
        nr_of_transferred = 1000
        i = 0
        transferring = []
        actions = []

        transfer_start = time.time()
        # transfer_info.log_stats(self.pg, self.schema, log_entry)
        for doc in docs:
            transferring.append(doc)
            try:
                if (i+1) % nr_of_transferred == 0 and i+1 >= nr_of_transferred:
                    if self.include_extra_props is True:
                        r.insert_bulk(
                            transferring,
                            self.attr_details,
                            self.include_extra_props)
                    else:
                        r.insert_bulk_no_extra_props(
                            transferring,
                            self.attr_details,
                            self.include_extra_props)
                    transferring = []
                if i + 1 == nr_of_docs and (i + 1) % nr_of_transferred != 0:
                    if self.include_extra_props is True:
                        r.insert_bulk(
                            transferring,
                            self.attr_details,
                            self.include_extra_props)
                    else:
                        r.insert_bulk_no_extra_props(
                            transferring,
                            self.attr_details,
                            self.include_extra_props)
                        logger.info(
                            "%s Finished collection %s: %d docs"
                            % (CURR_FILE,
                               coll, i + 1))
                        transferring = []
            except Exception as ex:
                logger.error("""%s Transfer unsuccessful. %s""" % (
                    CURR_FILE,
                    ex))
            if (i+1) % (nr_of_transferred * 10) == 0:
                logger.info("""%s %d/%d (%s)""" % (
                    CURR_FILE,
                    i+1,
                    nr_of_docs,
                    coll))
            i += 1

        transfer_end = time.time()
        action = 'UPSERT'
        if self.drop is True or self.truncate is True:
            action = 'INSERT'
        actions.append([action, transfer_start, transfer_end])

        vacuum_start = time.time()
        r.vacuum()
        vacuum_end = time.time()
        actions.append(['FULL VACUUM', vacuum_start, vacuum_end])

        log_entries = []

        for action in actions:
            log_entries.append(tuple([action[0], r.relation_name,
                                      nr_of_docs, action[1], action[2]]))
        transfer_info.log_stats(self.pg, self.schema, log_entries)

    def insert_multiple(self, docs, r, coll):
        '''
        Transfers multiple documents with different fields
        (not whole collections).
        Used by [TAILER]
        Parameters
        ----------
        doc : dict
            : document
        r : Relation
            relation in PG
        coll : string
             : collection name
        Returns
        -------
        -

        Raises
        ------
        Example
        -------
        '''
        (
            att_new,
            att_orig,
            types,
            name_rel,
            type_x_props_pg
        ) = cp.config_fields(self.coll_def, coll)

        # TODO: check if this is necessary:
        if types == []:
            return

        # Adding extra properties to inserted/updated row is necessary
        # because this attribute is not part of the original document
        # and anything that is not defined in the collection.yml file
        # will be pushed in this value. This function will also create
        # a dictionary which will contain all the information about
        # the attribute before and after the conversion.

        self.attr_details = self.prepare_attr_details(
            att_new, att_orig, types, type_x_props_pg)
        # TODO remove this stuff with the extra props
        try:
            if self.include_extra_props is True:
                r.insert_bulk(
                    docs, self.attr_details, self.include_extra_props)
            else:
                r.insert_bulk_no_extra_props_tailed(
                    docs, self.attr_details, self.include_extra_props)
        except Exception as ex:
            logger.error("""
            %s Transferring to %s was unsuccessful.
            Exception: %s
            """ % (
                CURR_FILE,

                r.relation_name, ex))
            logger.error("%s\n" % docs)

    def update_multiple(self, docs, r, coll):
        '''
        Upserts multiple documents with different fields.
        Parameters
        ----------
        doc : dict
            : document
        r : Relation
            relation in PG
        coll : string
             : collection name
        Returns
        -------
        -

        Raises
        ------
        Example
        -------
        '''
        (attrs_new, attrs_original, types, relation_name,
         type_x_props_pg) = cp.config_fields(self.coll_def, coll)
        if types == []:
            return
        # Adding extra properties to inserted/updated row is necessary
        # because this attribute is not part of the original
        # document and anything that is not defined in the
        # collection.yml file will be pushed in this value.
        # This function will also create a dictionary which will
        # contain all the information
        # about the attribute before and after the conversion.

        self.attr_details = self.prepare_attr_details(
            attrs_new, attrs_original, types, type_x_props_pg)
        # TODO remove this stuff with the extra props.
        try:
            if self.include_extra_props is True:
                r.insert_bulk(
                    docs, self.attr_details, self.include_extra_props)
            else:
                r.insert_bulk_no_extra_props_tailed(
                    docs, self.attr_details, self.include_extra_props)
        except Exception as ex:
            logger.error("""
            %s Transferring to %s was unsuccessful. Exception: %s
            """ % (
                CURR_FILE,
                r.relation_name, ex))
            logger.error('%s\n' % docs)

    def prepare_attr_details(self,
                             attrs_cm,
                             attrs_mdb,
                             types_cm,
                             type_x_props_pg=None):
        '''
        Adds extra properties field to the attribute details:
        (attr_details).
        Extra properties are not part of the original document
        and they need to be added in this separate step.
        Returns
        -------
        attr_details : list
                      : details for mapping each field
                      - name of the field (in mongodb)
                      - name in the collection map (for pg)
                      - type in the collection map (for pg)
                      - default value

        Parameters
        ----------
        attrs_cm : list
                    : attribute names from config file
        attrs_mdb : list
                    : field names of MongoDB document
        types_cm : list
                    : types from config files
        extra_props_type : string
                    : type of the extra property
        Example
        -------
        attrs_new = [kit_cat, birdy_bird]
        attrs_original = [kitCat, birdyBird]
        types = ['text', 'text']
        extra_props_type = 'jsonb'
        res = append_extra_props(
            attrs_new, attrs_original, types, extra_props_type
        )
        '''
        if self.include_extra_props is True:

            attrs_cm.append(name_extra_props_pg)
            attrs_mdb.append(name_extra_props_mdb)
            types_cm.append(type_x_props_pg)

        self.attr_details = {}
        for i in range(0, len(attrs_mdb)):
            details = {}
            details["name_cm"] = attrs_cm[i]
            details["type_cm"] = types_cm[i]
            details["value"] = None
            self.attr_details[attrs_mdb[i]] = details
        return self.attr_details

    def adjust_columns(self, coll):
        """
        Adds or removes extra properties if necessary and updates column types
        for a collection.
        Changes the attribute details (attr_details)

        Parameters
        ----------
        coll : string
             : name of the collection


        Returns
        ----------
        r : object
          : the relation object



        """
        # get data from the collection map
        (attrs_new,
         attrs_original,
         types,
         relation_name,
         type_x_props
         ) = cp.config_fields(self.coll_def, coll)
        if types == []:
            return

        r = relation.Relation(self.pg, self.schema, relation_name)

        self.add_extra_props(attrs_original, attrs_new, types)
        # Check if changing type was unsuccessful.
        failed = r.update_schema(self.attr_details)
        self.handle_failed_type_update(failed)

        return r

    def handle_failed_type_update(self, failed):
        if failed is not None:
            for tuf in failed:
                name_pg = tuf[0]
                name_mdb = [
                    attr for attr in self.attr_details if self.attr_details[attr]["name_cm"] == name_pg][0]
                type_orig = tuf[1].lower()
                type_new = self.attr_details[name_mdb]["type_cm"].lower()
                self.attr_details[name_mdb]["type_cm"] = type_orig
                logger.warn("""
                    %s Type conversion is not possible for column '%s'.
                    Skipping conversion %s -> %s.""" %
                            (
                                CURR_FILE,
                                name_pg, type_orig, type_new))

    def add_extra_props(self, attrs_original, attrs_new, types):

        # This dict contains all the necessary information about the
        # Mongo fields, Postgres columns and their types
        self.attr_details = {}
        attrs_mdb = attrs_original
        attrs_cm = attrs_new
        types_cm = types

        if self.include_extra_props is True:
            attrs_cm.append(name_extra_props_pg)
            types_cm.append(types)
            attrs_mdb.append(name_extra_props_mdb)

        for i in range(len(attrs_mdb)):
            details = {}
            details["name_cm"] = attrs_cm[i]
            details["type_cm"] = types[i]
            details["value"] = None
            self.attr_details[attrs_mdb[i]] = details



from etl.monitor import logger


def add_pk(db, schema, table, attr):
    """
    Adds primary key to a PostgreSQL table.

    Parameters
    ----------
    db : obj
    schema : string
    table : string
    attr  : string

    Returns
    -------
    -

    Example
    -------
    add_pk(pg, 'public', 'employee', 'id')

    """
    cmd = 'ALTER TABLE %s.%s ADD PRIMARY KEY (%s)' % (
        schema, table.lower(), attr)
    try:
        db.execute_cmd(cmd)
    except Exception as ex:
        logger.error(
            """
            [CONSTRAINT] Failed to add primary key to table %s.
            Details: %s
            """ % (table, ex))



from threading import Thread
from etl.extract import tailer, transfer_info
from etl.transform import config_parser as cp
from etl.load import schema
from datetime import datetime
from etl.monitor import logger


def start(extractor, coll_config):
    collections = cp.config_collection_names(coll_config)
    if collections is None:
        logger.error(
            """
                [TRANSFER] No collections found.
                Check your collection names in the setup file.
                """)
        return
    extractor.transfer(collections)


class TransferThread(Thread):
    new = False

    def __init__(self, settings, coll_config, pg, mongo, ex, tail, dt=None):
        Thread.__init__(self)
        self.settings = settings
        self.coll_config = coll_config
        self.pg = pg
        self.mongo = mongo
        self.t = tailer.Tailer(self.pg, self.mongo,
                               self.settings["postgres"],
                               self.settings,
                               self.coll_config)
        # initialize extractor
        self.extractor = ex
        self.terminate = False
        self.dt = dt
        self.start_tailing = tail

    def run(self):
        # collections which will be transferred
        if self.start_tailing is False:
            setup_pg = self.settings["postgres"]

            tailing_cmd = self.settings["tailing_from"]
            tailing_db = self.settings["tailing_from_db"]
            if tailing_db is False and tailing_cmd is None:

                schema.create(self.pg, setup_pg["schema_name"])

                # after extracting all collections tailing
                # will start from this timestamp
                if setup_pg["schema_reset"] is True:
                    schema.reset(self.pg, setup_pg["schema_name"])

                transfer_info.save_logs_to_db(self.pg, setup_pg["schema_name"])

                start(self.extractor, self.coll_config)
        else:
            self.tail(self.dt)

    def stop(self):
        self.terminate = True
        self.t.stop_tailing = True

    def tail(self, start_date_time=None):
        self.t.stop_tailing = False

        tailing_cmd = self.settings["tailing_from"]
        tailing_db = self.settings["tailing_from_db"]

        if self.settings["tailing"] is True:
            logger.info("[TRANSFER] Starting standard tailing.")
            self.t.start(start_date_time)

        elif tailing_cmd is not None:
            logger.info("[TRANSFER] Starting tailing from provided timestamp.")
            self.t.start(tailing_cmd)

        elif tailing_db is True or start_date_time is None:
            logger.info(
                """[TRANSFER] Starting tailing from
                timestamp found in purr_info.""")
            ts = transfer_info.get_latest_successful_ts(self.pg, 'public')
            latest_ts = int((list(ts)[0])[0])
            self.t.start(latest_ts)
        else:
            self.pg.__del__()

    def tailing_stop(self):
        self.t.stop_tailing = True

    def schema_update(self):
        self.extractor.update_coll_map()



import time

from etl.load import table, row
from etl.monitor import logger

table_desc = {
    "purr_oplog": {
        "attrs": ["id", "operation", "relation", "obj_id", "ts", "merged", "document"],
        "types": ["SERIAL", "TEXT", "TEXT", "TEXT", "INTEGER", "BOOLEAN", "TEXT"],
        "pks": ["id", "ts"]
    },
    "purr_info": {
        "attrs": ["id", "relation", "latest_successful_ts"],
        "types": ["INTEGER", "TEXT", "TEXT"]
    },
    "purr_transfer_stats": {
        "attrs": ["id", "action", "relation", "number_of_rows", "ts_start", "ts_end"],
        "types": ["SERIAL", "TEXT", "TEXT", "INTEGER", "INTEGER", "INTEGER"]
    },
    "purr_error": {
        "attrs": ["id", "location", "message", "ts"],
        "types": ["SERIAL", "TEXT", "TEXT", "INTEGER"]
    }
}

def save_logs_to_db(db, schema='public'):
    create_stat_table(db, schema)
    create_oplog_table(db, schema)
    create_transfer_stats_table(db, schema)
    create_log_error_table(db, schema)

def create_stat_table(db, schema='public'):
    """
    Creates a table that holds the timestamp of the
    latest successfully inserted item.
    Parameters
    ----------

    Returns
    -------
    -

    Example
    -------
    create_stat_table(pg, 'purr')

    """
    table_name = "purr_info"
    attrs = table_desc[table_name]["attrs"]
    types = table_desc[table_name]["types"]
    values = [0, None, int(time.time())]
    try:
        table.create(db, schema, table_name, attrs, types)
        ts = get_latest_successful_ts(db, schema)
        if len(ts) == 0:
            row.insert(db, schema, table_name, attrs, values)
        logger.info("[TRANSFER INFO] Created table %s." % (table_name))
    except Exception as ex:
        logger.error(
            "[TRANSFER_INFO] Failed to create table %s: %s" % (table_name, ex))


def get_latest_successful_ts(db, schema='public'):
    """
  Get the timestamp of the latest successful transfer.

  Parameters
  ----------

  Returns
  -------
  -

  Example
  -------
  get_latest_successful_ts(pg, 'purr')

  """
    table_name = 'purr_info'
    cmd = "SELECT latest_successful_ts FROM %s.%s;" % (schema, table_name)

    try:
        res = db.execute_cmd_with_fetch(cmd)
        return res
    except Exception as ex:
        logger.error(
            """[TRANSFER_INFO] Failed to get the timestamp
             of the latest successful transfer: %s"""
            % (ex)
        )


def update_latest_successful_ts(db, schema, dt):
    cmd = "UPDATE %s.purr_info SET latest_successful_ts='%s';" % (
        schema, str(dt))
    try:
        db.execute_cmd(cmd)
    except Exception as ex:
        logger.error(
            """[TRANSFER_INFO] Failed to update the timestamp
            of the latest successful transfer: %s"""
            % (ex)
        )


def create_oplog_table(db, schema='public'):
    """
    Logs the operation, relation name, object id and
    timestamp for each entry of the oplog.

    Parameters
    ----------
    db: connection obj
    schema: name of the schema in Postgres
    Returns
    -------
    -

    Example
    -------
    create_oplog_table(pg, 'purr')

    """
    table_name = "purr_oplog"
    attrs = table_desc[table_name]["attrs"]
    types = table_desc[table_name]["types"]
    pks = table_desc[table_name]["pks"]

    values = [int(time.time())]
    try:
        table.drop(db, schema, [table_name])
        table.create(db, schema, table_name, attrs, types, pks)
        logger.info("[TRANSFER INFO] Created table %s." % (table_name))
    except Exception as ex:
        logger.error(
            "[TRANSFER_INFO] Failed to create table %s: %s" % (table_name, ex))


def log_rows(db, schema, values):
    """
    Holds the operation, relation name, object id and
    timestamp for each entry of the oplog.

    Parameters
    ----------

    Returns
    -------
    -

    Example
    -------
    create_oplog_table(pg, 'purr')

    """
    table_name = "purr_oplog"
    # id is SERIAL type, we can skip it when inserting rows:
    attrs = table_desc[table_name]["attrs"][1:]
    try:
        row.insert_bulk(db, schema, table_name, attrs, values)
    except Exception as ex:
        logger.error(
            "[TRANSFER_INFO] Failed to insert logs into table %s: %s"
            % (table_name, ex))


def create_transfer_stats_table(db, schema='public'):
    """
    Logs the number, relation name, timestamp 
    for each collection transfer.

    Parameters
    ----------
    db: connection obj
    schema: name of the schema in Postgres
    Returns
    -------
    -

    Example
    -------
    create_transfer_stats_table(pg, 'purr')

    """
    table_name = "purr_transfer_stats"
    attrs = table_desc[table_name]["attrs"]
    types = table_desc[table_name]["types"]

    values = [int(time.time())]
    try:
        table.create(db, schema, table_name, attrs, types)
        logger.info("[TRANSFER INFO] Created table %s." % (table_name))
    except Exception as ex:
        logger.error(
            "[TRANSFER_INFO] Failed to create table %s: %s" % (table_name, ex))


def log_stats(db, schema, values):
    """
    Insert the number, relation name, timestamp 
    for each collection transfer.

    Parameters
    ----------

    Returns
    -------
    -

    Example
    -------
    log_stats(pg, 'purr', [])

    """
    table_name = "purr_transfer_stats"
    # id is SERIAL type, we can skip it when inserting rows:
    attrs = table_desc[table_name]["attrs"][1:]
    try:
        row.insert_bulk(db, schema, table_name, attrs, values)
    except Exception as ex:
        logger.error(
            "[TRANSFER_INFO] Failed to insert logs into table %s: %s"
            % (table_name, ex))


def create_log_error_table(db, schema='public'):
    """
    Logs the error's location, message and timestamp 
    when an it occurs.

    Parameters
    ----------
    db: connection obj
    schema: name of the schema in Postgres
    Returns
    -------
    -

    Example
    -------
    create_log_error_table(pg, 'purr')

    """
    table_name = "purr_error"
    attrs = table_desc[table_name]["attrs"]
    types = table_desc[table_name]["types"]

    values = [int(time.time())]
    try:
        table.create(db, schema, table_name, attrs, types)
        logger.info("[TRANSFER INFO] Created table %s." % (table_name))
    except Exception as ex:
        logger.error(
            "[TRANSFER_INFO] Failed to create table %s: %s" % (table_name, ex))


def log_error(db, values, schema='public'):
    """
    Insert the number, relation name, timestamp 
    for each collection transfer.

    Parameters
    ----------

    Returns
    -------
    -

    Example
    -------
    log_stats(pg, 'purr', [])
    """
    
    table_name = "purr_error"
    # id is SERIAL type, we can skip it when inserting rows:
    attrs = table_desc[table_name]["attrs"][1:]
    try:
        row.insert(db, schema, table_name, attrs, values)
    except Exception as ex:
        logger.error("""[TRANSFER_INFO] Failed to insert logs into table %s: %s"""
                     % (table_name, ex))



import json


def cast(value, column_type):
    '''
    column_type: string
    value: string
    Casts value based on required column type in collections.yml.
    When serializing ObjectId and datetime, a string is returned
    instead of objects with $oid and $date keys.
    Example:
      { _id: { $oid: "56ac26cc05b37b4e1a4d2222" } }
      { _id: "56ac26cc05b37b4e1a4d2222" }

    Parameters:
    -----------
    value       : anything
                  value to cast
    column_type : string
                  type of column
    '''
    new_value = None

    if value is None:
        return new_value

    if value == '$unset':
        return value

    column_type = column_type.lower()

    if column_type == 'text':
        new_value = str(value)

    elif column_type == 'timestamp':
        new_value = value

    elif column_type == 'double precision':
        try:
            new_value = float(str(value))
        except ValueError:
            new_value = 'undefined'

    elif column_type == 'jsonb':
        new_value = json.dumps(value, default=str)

    elif column_type == 'boolean':
        new_value = bool(value)

    return new_value



import argparse
import sys
import time
from datetime import datetime

import etl.core
from etl.transform import config_parser

setup = {}


def valid_date(s):
    try:
        t = datetime.strptime(s, "%d-%m-%yT%H:%M:%S")
        epoch_time = int(time.mktime(t.timetuple()))
        return epoch_time
    except ValueError:
        msg = "Not a valid date: '{0}'.".format(s)
        raise argparse.ArgumentTypeError(msg)


def main():
    parser = argparse.ArgumentParser(
        description="""
        An ETL tool to transfer data from MongoDB to Postgres.
        """
    )

    parser.add_argument('-sf', '--setup-file', type=str,
                        dest="path_setup_file",
                        help="""
                        Path to YAML file bith basic setup: \n
                        connection string for MongoDB and PG,
                        schema name, schema reset, collections to
                        transfer, etc.""")
    parser.add_argument('-cf', '--collection-file', type=str,
                        dest='path_collection_file',
                        help="""
                        Path to YAML file that contains information about
                        collections: collection names, field names, types,
                        PG relation and attribute names, etc.
                        """)
    parser.add_argument('-td', '--table-drop', action='store_true',
                        dest='table_drop', default=False, help='')
    parser.add_argument('-tt', '--table-truncate', action='store_true',
                        dest='table_truncate', default=False, help='')
    parser.add_argument('-sr', '--schema-reset', action='store_true',
                        dest='schema_reset', default=False, help='')
    parser.add_argument('-sn', '--schema-name', type=str,
                        dest='schema_name', default='public', help='')
    parser.add_argument('-pg', '--pg-connection', type=str,
                        dest='pg_connection', help='')
    parser.add_argument('-mdb', '--mongo-connection', type=str,
                        dest='mongo_connection', default='', help='')
    parser.add_argument('-n', '--mongo-db-name', type=str,
                        dest='mongo_db_name', default='', help='')
    parser.add_argument('-t', '--tail', action='store_true',
                        dest='tail', default=False, help='')
    parser.add_argument('-m', '--map', action='store_true',
                        dest='map', help='Generate a collection map')
    parser.add_argument("-s",
                        "--start",
                        dest="timestamp",
                        nargs='?',
                        help="""Start tailing from specific timestamp; \n
                            use in combination with --tail in order to
                            tail the oplog from a specific date and time.
                            Format: DD-MM-YYThh:mm:ss.
                            Example: 25-07-18T16:30:00""",
                        type=valid_date)
    parser.add_argument('-tsdb', '--start-from-tsdb', action='store_true',
                        dest='timestamp_db', default=False,
                        help="""
                        Start tailing from latest saved timestamp.
                        Looks for the value saved in table purr_info.
                        """)

    parser.add_argument('-ta', '--typecheck-auto', action='store_true',
                        dest='typecheck_auto', default=False, help='')
    parser.add_argument('-ex', '--include-extra-properties',
                        action='store_true',
                        dest='include_extra_props',
                        default=False, help='')

    args = parser.parse_args()

    if len(sys.argv) <= 1:
        sys.argv.append('--help')

    setup_file = {}
    coll_file = {}
    if args.map is True and args.mongo_connection and args.mongo_db_name:
        etl.core.generate_collection_map({
            'connection': args.mongo_connection,
            'db_name': args.mongo_db_name
        })
    else:
        if args.path_collection_file:
            coll_file = config_parser.config_collections(
                args.path_collection_file)
        if args.pg_connection and args.mongo_connection and args.mongo_db_name:
            setup = {
                'postgres':
                {
                    'connection': args.pg_connection,
                    'schema_name': args.schema_name,
                    'schema_reset': args.schema_reset,
                    'table_truncate': args.table_truncate,
                    'table_drop': args.table_drop
                },
                'mongo':
                {
                    'connection': args.mongo_connection,
                    'db_name': args.mongo_db_name
                },
                'tailing': args.tail,
                'tailing_from': args.timestamp,
                'tailing_from_db': args.timestamp_db,
                'typecheck_auto': args.typecheck_auto,
                'include_extra_props': args.include_extra_props
            }

            # TODO this is bad ↓ merge tailing
            tailing_from_ts = setup["tailing_from"] or setup["tailing_from_db"]
            if setup["tailing"] and tailing_from_ts:
                print("\nOnly one tailing option is allowed.")
                raise SystemExit()
            etl.core.start(setup, coll_file)

        elif args.path_setup_file:
            setup_file = config_parser.config_basic(args.path_setup_file)
            if setup_file and coll_file:
                etl.core.start(setup_file, coll_file)
        else:
            print("Check your input...")



from datetime import datetime
import time
import os
from etl.extract import extractor, transfer
from etl.extract import init_mongo as mongodb, notification
from etl.extract import collection_map as cm
from etl.load import init_pg as postgres
from etl.monitor import logger
import pkg_resources

global THREADS
THREADS = []

CURR_FILE = "[CORE]"


def start(settings, coll_map):
    """
    Starts Purr.
    Returns
    -------
    -

    Parameters
    ----------
    settings : dict
             : basic settings for both PG and MongoDB
             (connection strings, schema name)

    coll_map : dict
                : config file for collections

    TODO
    ----
    - create table with attributes and types
    """

    logger.info("Starting Purr v%s ..." %
                pkg_resources.require("purrito")[0].version, CURR_FILE)

    logger.info("PID=%s" % os.getpid())

    mode_tailing = False
    if settings["tailing"] or settings["tailing_from"] or settings["tailing_from"]:
        mode_tailing = True

    logger.info("TAILING=%s" % ("ON" if mode_tailing else "OFF"))

    setup_pg = settings["postgres"]
    setup_mdb = settings["mongo"]

    pg = postgres.PgConnection(setup_pg["connection"])
    mongo = mongodb.MongoConnection(setup_mdb)

    cm.create_table(
        pg, coll_map, setup_pg["schema_name"])

    ex = extractor.Extractor(
        pg, mongo.conn, setup_pg, settings, coll_map)

    start_date_time = datetime.utcnow()

    if mode_tailing:
        try:
            # first just transfer the data
            tailing_start = False
            thr_transfer = transfer.TransferThread(
                settings,
                coll_map,
                pg,
                mongo,
                ex,
                tailing_start
            )
            thr_transfer.start()
            # wait until thread is finished
            thr_transfer.join()
            handle_coll_map_changes(settings, coll_map, pg,
                                    mongo, ex, start_date_time)
        except (KeyboardInterrupt, SystemExit):
            logger.error('Stopping transfer.', CURR_FILE)
            for t in THREADS:
                t.stop()
            raise SystemExit()

        except Exception as ex:
            logger.error(
                "Unable to start transfer thread. Details: %s" % ex, CURR_FILE)
            raise SystemExit()

    else:
        transfer.start(ex, coll_map)


def handle_coll_map_changes(settings, coll_map, pg, mongo, ex, dt):
    tailing_start = True
    try:
        thr_notification = notification.NotificationThread(pg)
        THREADS.append(thr_notification)
        thr_notification.start()

        thr_transfer = transfer.TransferThread(
            settings, coll_map, pg, mongo, ex, tailing_start, dt)
        THREADS.append(thr_transfer)
        thr_transfer.start()
        while True:
            time.sleep(2)
            if thr_notification.new is True:
                thr_transfer.tailing_stop()

                thr_notification.new = False
                thr_transfer.join(3)

                thr_transfer = transfer.TransferThread(
                    settings, coll_map, pg, mongo, ex, tailing_start, dt)
                THREADS.append(thr_transfer)
                thr_transfer.settings["tailing"] = False
                thr_transfer.settings["tailing_from_db"] = True
                thr_transfer.schema_update()
                thr_transfer.start()
        thr_notification.join(3)  # wait for 3 seconds
        thr_transfer.join(3)
    except (KeyboardInterrupt, SystemExit):
        logger.error('Stopping all threads.', CURR_FILE)
        for t in THREADS:
            t.stop()
        raise SystemExit()
    except Exception as ex:
        logger.error(
            "Unable to start listener thread. Details: %s" % ex, CURR_FILE)
        raise SystemExit()


def generate_collection_map(settings_mdb):
    """
    TODO:
    - add docs
    - disconnect from Mongo!
    """
    logger.info("Starting Purr v%s ..." %
                pkg_resources.require("purrito")[0].version, CURR_FILE)

    logger.info("PID=%s" % os.getpid(), CURR_FILE)
    mongo = mongodb.MongoConnection(settings_mdb)
    coll_map = cm.create_map(mongo.conn, settings_mdb["db_name"])
    cm.create_file(coll_map)
    mongo.disconnect()



import psycopg2
import psycopg2.extras
import time
from etl.monitor import logger
from etl.extract import transfer_info
from psycopg2.extensions import ISOLATION_LEVEL_AUTOCOMMIT
from enum import Enum

CURR_FILE = "[INIT_PG]"


class FunctionLatest(Enum):
    EXECUTE = 1,
    EXECUTE_WITH_FETCH = 2,
    EXECUTE_MANY = 3


class PgConnection:

    def __init__(self, conn_details, ttw=1):
        logger.info("%s Connecting to %s" % (CURR_FILE, conn_details))

        # time to wait before attempt to reconnect
        self.ttw = ttw
        self.conn_details = conn_details
        self.cmd_latest = None
        self.values_latest = None
        self.function_latest = None
        self.query_failed = False

        if ttw == 1:
            self.attempt_to_reconnect = False
        try:
            self.conn = psycopg2.connect(self.conn_details)
            self.conn.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)

            self.cur = self.conn.cursor()
            logger.info("%s Connected to Postgres." % (CURR_FILE))
            self.ttw = 1
            if self.query_failed is True:
                """
                the latest command should be repeated because Postgres
                was disconnected and it is likely that it was during a
                command
                """
                self.query_failed = False
                if self.function_latest is FunctionLatest.EXECUTE:
                    self.execute_cmd(self.cmd_latest, self.values_latest)
                elif self.function_latest is FunctionLatest.EXECUTE_WITH_FETCH:
                    self.execute_cmd_with_fetch(
                        self.cmd_latest, self.values_latest)
                elif self.function_latest is FunctionLatest.EXECUTE_MANY:
                    self.execute_many_cmd(self.cmd_latest, self.values_latest)

        except Exception as ex:
            self.attempt_to_reconnect = True
            msg = """
                %s Could not connect to Postgres.
                Reconnecting in %s seconds...
                Details: %s
                """ % (CURR_FILE, self.ttw, ex)
            self.log_error_in_pg(msg)
            time.sleep(self.ttw)

            self.__init__(self.conn_details, self.ttw * 2)

    def execute_cmd(self, cmd, values=None):
        self.cmd_latest = cmd
        self.values_latest = values
        self.function_latest = FunctionLatest.EXECUTE

        cur = self.conn.cursor()
        try:
            if values is not None:
                cur.execute(cmd, values)
            else:
                cur.execute(cmd)

        except (psycopg2.InterfaceError, psycopg2.OperationalError) as ex:
            self.handle_interface_and_oper_error()
            self.log_error_in_pg(ex)
            self.query_failed = True
        except Exception as ex:
            self.query_failed = True
            msg = """
                %s Executing query without fetch failed.
                Details: %s
                """ % (CURR_FILE, ex)
            self.log_error_in_pg(msg)
        cur.close()

    def execute_cmd_with_fetch(self, cmd, values=None):
        self.cmd_latest = cmd
        self.values_latest = values
        self.function_latest = FunctionLatest.EXECUTE_WITH_FETCH

        cur = self.conn.cursor()

        try:
            if values is not None:
                cur.execute(cmd, values)
            else:
                cur.execute(cmd)
            return cur.fetchall()

        except (psycopg2.InterfaceError, psycopg2.OperationalError):
            self.handle_interface_and_oper_error()
        except Exception as ex:
            self.query_failed = True
            msg = """
                %s Executing query with fetch failed.
                Details: %s
                """ % (CURR_FILE, ex)
            self.log_error_in_pg(msg)

        cur.close()

    def execute_many_cmd(self, cmd, values):
        self.cmd_latest = cmd
        self.values_latest = values
        self.function_latest = FunctionLatest.EXECUTE_MANY

        cur = self.conn.cursor()
        try:
            cur.executemany(cmd, values)
        except (psycopg2.InterfaceError, psycopg2.OperationalError):
            self.handle_interface_and_oper_error()
        except Exception as ex:
            self.query_failed = True
            msg = """
                %s Executing many failed.
                Details: %s
                """ % (CURR_FILE, ex)
            self.log_error_in_pg(msg)
        cur.close()

    def poll(self):
        cur = self.conn.cursor()
        cur.execute()
        cur.close()

    def notifies(self):
        return self.conn.notifies

    def __del__(self):
        self.conn.close()
        self.cur.close()

    def handle_interface_and_oper_error(self):
        msg = "%s Trying to reconnect to Postgres..." % (CURR_FILE)
        self.log_error_in_pg(msg)
        self.attempt_to_reconnect = True
        self.__init__(self.conn_details, self.ttw * 2)

    def log_error_in_pg(self, msg):
        logger.error(msg)
        msg_full = """Message: %s\nCommand: %s\nValues: %s""" % (
            msg, self.cmd_latest, self.values_latest)
        row = tuple(['PG', msg_full, time.time()])
        transfer_info.log_error(self, row)





def test_create_stat_table():
    pass


def test_get_latest_successful_ts():
    pass


def test_update_latest_successful_ts():
    pass


def test_create_oplog_table():
    pass


def test_log_rows():
    pass



from etl.load import table, row, constraint
from etl.transform import type_checker
from etl.transform import unnester


def init_values(attrs):
    """
    Sets value for every column to None
    """
    for k, v in attrs.items():
        attrs[k]["value"] = None
    return attrs


def set_values(attr_details, doc, _extra_props=None):
    """
    Casts values for a whole document
    """
    for key, field_value in doc.items():
        keys_cm = list(attr_details.keys())
        if key in keys_cm:
            field_type = attr_details[key]["type_cm"]
            value = unnester.cast(
                field_value, field_type)
            if value == 'undefined' and _extra_props is not None:
                _extra_props.update({key: field_value})
            else:
                attr_details[key]["value"] = value
        else:
            if _extra_props is not None:
                _extra_props.update({key: field_value})
    if _extra_props is not None:
        return attr_details, _extra_props
    else:
        return attr_details


def prepare_row_for_insert(attrs, doc, include_extra_props=None):
    _extra_props = {}
    attrs = init_values(attrs)

    (attrs, _extra_props) = set_values(attrs, doc, _extra_props)

    if include_extra_props is True:
        _extra_props = unnester.cast(_extra_props, 'jsonb')
        attrs["extraProps"]["value"] = _extra_props

    attrs_pg = [v["name_cm"]
                for k, v in attrs.items() if k in doc.keys()]
    values = [v["value"] for k, v in attrs.items() if k in doc.keys()]
    if len(doc.keys()) > len(attrs_pg) and include_extra_props:
        attrs_pg.append('_extra_props')
        values.append(_extra_props)
    return attrs_pg, values


def is_schema_changed(attrs_pg, types_pg, attrs_cm, types_cm):
    equal_attrs = set(attrs_cm) == set(attrs_pg)
    equal_types = set(types_cm) == set(types_pg)
    if len(attrs_pg) and (equal_attrs and equal_types):
        return False
    return True


class Relation():
    """
    This is the main parents class for transforming data.
    """

    def __init__(self, pg, schema, relation, created=False):
        """Constructor for Relation"""
        self.relation_name = relation
        self.created = created
        self.db = pg
        self.schema = schema

    def exists(self):
        self.created = table.exists(self.db, self.schema, self.relation_name)
        return self.created

    def insert_bulk(self, docs, attrs,
                    include_extra_props=True):
        """
        Transforms document and inserts it into the corresponding table.
        Parameters
        ----------
        doc : dict
              the document we want to insert
        TODO add unset
        """
        # This is needed because
        # sometimes there is no value for attributes (null)
        result = []
        if type(docs) is not list:
            docs = [docs]
        for doc in docs:
            (attrs_pg, values) = prepare_row_for_insert(attrs, doc)
            result.append(tuple(values))

        if self.created is True:
            row.upsert_bulk(self.db, self.schema,
                            self.relation_name, attrs_pg, result)
        else:
            row.insert_bulk(self.db, self.schema,
                            self.relation_name, attrs_pg, result)

    def insert_bulk_no_extra_props(self, docs, attrs,
                                   include_extra_props=True):
        """
        Transforms document and inserts it into the corresponding table.
        Parameters
        ----------
        docs : dict
              the documents we want to insert
         unset: string[]
              list of fields to unset
        """
        # This is needed because
        # sometimes there is no value for attributes (null)
        result = []
        if type(docs) is not list:
            docs = [docs]
        for doc in docs:
            attrs = init_values(attrs)
            attrs = set_values(attrs, doc)

            if len(docs) > 1:
                attrs_pg = [v["name_cm"] for k, v in attrs.items()]
                values = [v["value"] for k, v in attrs.items()]
            else:
                attrs_pg = [v["name_cm"]
                            for k, v in attrs.items() if k in doc.keys()]
                values = [v["value"]
                          for k, v in attrs.items() if k in doc.keys()]
            result.append(tuple(values))

        if self.created is True or len(docs) == 1:
            row.upsert_bulk(self.db, self.schema,
                            self.relation_name, attrs_pg, result)
        else:
            row.insert_bulk(self.db, self.schema,
                            self.relation_name, attrs_pg, result)

    def insert_bulk_no_extra_props_tailed(self,
                                          docs,
                                          attrs,
                                          include_extra_props=True):
        """
        Transforms document and inserts it into the corresponding table.
        Parameters
        ----------
        docs : dict
              the documents we want to insert

        """
        # This is needed because
        # sometimes there is no value for attributes (null)
        result = []
        if type(docs) is not list:
            docs = [docs]
        grouped_values = {}

        for i in range(len(docs)):
            attrs_pg = []
            values = []
            doc = docs[i]
            attrs = init_values(attrs)
            (attrs) = set_values(
                attrs, doc)

            attrs_pg = tuple([v["name_cm"]
                              for k, v in attrs.items() if k in doc.keys()])
            values = [v["value"]
                      for k, v in attrs.items() if k in doc.keys()]

            if attrs_pg not in grouped_values.keys():
                grouped_values[attrs_pg] = []
            grouped_values[attrs_pg].append(values)

        for k, v in grouped_values.items():
            attrs_pg = list(k)
            row.upsert_bulk_tail(self.db, self.schema,
                                 self.relation_name, attrs_pg, v)

    def delete(self, docs):
        ids = []
        if type(docs) is list:
            for doc in docs:
                ids.append(str(doc["_id"]))
        else:
            ids.append(str(docs["_id"]))

        row.delete(self.db, self.schema, self.relation_name, ids)

    def create_with_columns(self, attrs, types):
        table.create(self.db, self.schema, self.relation_name, attrs, types)

    def add_pk(self, attr):
        constraint.add_pk(self.db, self.schema, self.relation_name, attr)

    def columns_remove(self, attrs_pg, types_pg, attrs_cm, types_cm):
        """
        Check if attributes in Postgres exist in the collection map.
        If not, remove the column.
        """
        temp_attrs_cm = []
        temp_types_cm = []

        for i in range(0, len(attrs_pg)):
            try:
                # check if attributes in PG are part of the collection map)
                idx = attrs_cm.index(attrs_pg[i])
            except ValueError:
                # remove extra columns from PG (because they are no longer
                # part of the collection map)
                table.remove_column(
                    self.db, self.schema, self.relation_name, attrs_pg[i])
                attrs_pg[i] = None
                types_pg[i] = None
                continue
            temp_attrs_cm.append(attrs_cm[idx])
            temp_types_cm.append(types_cm[idx])
            del attrs_cm[idx]
            del types_cm[idx]

        attrs_pg = [x for x in attrs_pg if x is not None]
        types_pg = [x for x in types_pg if x is not None]

        attrs_cm = temp_attrs_cm + (attrs_cm)
        types_cm = temp_types_cm + (types_cm)

        return attrs_pg, types_pg, attrs_cm, types_cm

    def columns_add(self, attrs_pg, types_pg, attrs_cm, types_cm):
        """
        Adds columns to Postgres OR updates column types.
        """
        type_convert_fail = []

        if set(attrs_cm).issubset(set(attrs_pg)):
            # update types
            for i in range(len(attrs_pg)):
                if attrs_pg[i] in attrs_cm:
                    # type from the db and type from the config file
                    type_old = types_pg[i].lower()
                    type_new = types_cm[i].lower()

                    name_ts_no_tz = 'timestamp without time zone'
                    name_ts = 'timestamp'
                    if type_old == name_ts_no_tz and type_new == name_ts:
                        continue
                    elif type_old != type_new:
                        # type was changed
                        if type_checker.is_convertable(type_old, type_new):
                            table.column_change_type(
                                self.db,
                                self.schema,
                                self.relation_name,
                                attrs_pg[i],
                                type_new)
                        else:
                            type_convert_fail.append(
                                (attrs_pg[i], type_old))
                            continue
            return type_convert_fail
        else:
            # add new columns
            diff = list(set(attrs_cm) - set(attrs_pg))

            # get type of new attributes
            attrs_to_add = []
            types_to_add = []
            for d in diff:
                attrs_to_add.append(d)
                idx = attrs_cm.index(d)
                types_to_add.append(types_cm[idx])
            table.add_multiple_columns(
                self.db,
                self.schema,
                self.relation_name,
                attrs_to_add,
                types_to_add
            )

    def update_schema(self, attrs_types_cm):
        """
        Checks if there were any changes in the schema and adds/changes
        attributes if needed.
        If relation is already created in the database then we need
        to get the existing attributes and types and compare them
        to our new attributes and types from the config file.

        - find what's different in attributes from pg and conf
        - check existing attribute names
        - check types
        - check check is one is castable to the other
        """

        attrs_cm = []
        types_cm = []
        attrs_pg = []
        types_pg = []
        column_info = table.get_column_names_and_types(
            self.db, self.schema, self.relation_name)
        if self.exists() is True and column_info is not None:
            # Every attribute from pg and conf has to have the same order.
            # We are sorting by pg column names.
            attrs_types_pg = dict(column_info)
            attrs_pg = [k for k in sorted(attrs_types_pg.keys())]
            types_pg = [attrs_types_pg[k]
                        for k in sorted(attrs_types_pg.keys())]

            attrs_cm = [attrs_types_cm[k]["name_cm"]
                        for k in attrs_types_cm.keys()]
            types_cm = [attrs_types_cm[k]["type_cm"]
                        for k in attrs_types_cm.keys()]

            # if attributes from PG and the collection map are
            # the same, do nothing
            schema_changed = is_schema_changed(
                attrs_pg, types_pg, attrs_cm, types_cm)

            if schema_changed is False:
                return

            (attrs_pg, types_pg, attrs_cm, types_cm) = self.columns_remove(
                attrs_pg, types_pg, attrs_cm, types_cm)

            self.columns_add(attrs_pg, types_pg, attrs_cm, types_cm)
        else:
            attrs_cm = [v["name_cm"] for k, v in attrs_types_cm.items()]
            types_cm = [v["type_cm"] for k, v in attrs_types_cm.items()]
            if self.exists() is False:
                self.create_with_columns(attrs_cm, types_cm)
            return
        # TODO if table was dropped or schema was reset
        # then there is no need to have fun
        # with the type checking.
        # if len(attrs_types_from_db) == 0:

        # When new attributes are fully contained in the attribute list from DB
        # we need to check if the types are equal and if not,
        # we need to check if it is possible to convert the
        # old type into the new one.
        # Anything can be converted to JSONB.

    def vacuum(self):
        table.vacuum(self.db, self.schema, self.relation_name)



import psycopg2
from etl.monitor import logger


def create(db, schema, name, attrs, types, pks=["id"]):
    """
    Creates a table in Postgres.
    Parameters
    ----------
    name : str
    TODO
    ----
    """
    attrs_and_types = []

    for i in range(len(attrs)):
        pair = '"%s" %s' % (attrs[i], types[i])
        attrs_and_types.append(pair)

    pks = [('"%s"' % pk) for pk in pks]
    primary_keys = "PRIMARY KEY (%s)" % ",".join(pks)
    attrs_and_types.append(primary_keys)
    attrs_and_types = ", ".join(attrs_and_types)

    name = name.lower()
    cmd = "CREATE TABLE IF NOT EXISTS %s.%s(%s);" % (
        schema, name, attrs_and_types)
    try:
        db.execute_cmd(cmd)
    except Exception as ex:
        logger.error('[TABLE] %s when executing command %s.' % (ex, cmd))


def exists(db, schema, table):
    """
    Check if a table exists in the PG database.

    Parameters
    ----------
    table : string

    Returns
    -------
    True: table exists in the database
    False: otherwise
    """
    cmd = """
    SELECT table_name FROM information_schema.tables
    WHERE table_schema='%s' AND table_name='%s';
    """ % (
        schema, table.lower())

    try:
        res = db.execute_cmd_with_fetch(cmd)
        if res:
            return True
        else:
            return False
    except Exception as ex:
        logger.error('[TABLE] %s when executing command %s.' % (ex, cmd))


def truncate(db, schema, tables):
    """
    Parameters
    ----------
    tables: string[]
    Deletes data from 1..* tables.
    TODO
    ----
    Check if table exists before doing anything.
    """
    tables_cmd = []
    for t in tables:
        tables_cmd.append('%s.%s' % (schema, t))
    tables_cmd = ','.join(tables_cmd)
    cmd = "TRUNCATE TABLE %s;" % (tables_cmd)

    try:
        db.execute_cmd(cmd)
    except Exception as ex:
        logger.error('[TABLE] %s when executing command %s.' % (ex, cmd))


def drop(db, schema, tables):
    """
    Drop one or more tables in the PG database.

    Parameters
    ----------
    schema : string
    tables : list

    Example
    -------
    drop(pg, 'public', ['my_table'])

    Todo
    ----
    - first check if all tables in the list exist
    """
    tables_cmd = []
    for t in tables:
        tables_cmd.append('%s.%s' % (schema, t.lower()))
    tables_cmd = ', '.join(tables_cmd)

    cmd = "DROP TABLE IF EXISTS %s" % (tables_cmd)
    try:
        db.execute_cmd(cmd)
        logger.info('[TABLE] Dropping table(s) %s.' % (tables_cmd))
    except Exception as ex:
        logger.error('[TABLE] %s when executing command %s.' % (ex, cmd))


def add_column(db, schema, table, column_name, column_type):
    """
    Add new column to a specific table.
    Parameters
    ----------
    table : str
          : name of table to alter
    column_name : str
                : name of new column
    column_type : str
                : type of new column
    Example
    -------
    add_column(db, 'some_integer', 'integer')
    """
    cmd = "ALTER TABLE IF EXISTS %s.%s ADD COLUMN IF NOT EXISTS %s %s;" % (
        schema, table.lower(), column_name, column_type)
    logger.warn("""
    [TABLE] Adding new column to table: %s, column: %s, type: %s
    """ % (
        table.lower(), column_name, column_type))
    try:
        db.execute_cmd(cmd)
    except Exception as ex:
        logger.error('[TABLE] %s when executing command %s.' % (ex, cmd))


def add_multiple_columns(db, schema, table, attrs, types):
    """
    Add new column to a specific table.
    Parameters
    ----------
    name : str
    column_name : str
    column_type : str

    Example
    -------
    add_multiple_columns(db, ['nyanya', some_integer'], ['text', integer'])
    """
    statements_add = []
    attrs_types = zip(attrs, types)
    for i, j in attrs_types:
        statements_add.append(' '.join(['ADD COLUMN IF NOT EXISTS', i, j]))
    statements_merged = ', '.join(statements_add)

    cmd = "ALTER TABLE IF EXISTS %s.%s %s;" % (
        schema, table.lower(), statements_merged)

    for i, j in zip(attrs, types):
        logger.warn("Adding column %s (%s) to company %s." %
                    (i, j, table.lower()))
    try:
        db.execute_cmd(cmd)
    except Exception as ex:
        logger.error('[TABLE] %s when executing command %s.' % (ex, cmd))


def column_change_type(db, schema, table, column_name, column_type):
    """
    Add new column to a specific table.
    Parameters
    ----------
    name : str
    column_name : str
    column_type : str

    Example
    -------
    column_change_type(pg.db, 'some_integer', 'integer')
    """
    expression = ''
    if column_type == 'jsonb':
        expression = 'to_json(%s)' % column_name
    elif column_type == 'double precision':
        expression = 'CAST(%s as double precision)' % column_name

    if len(expression) == 0:
        cmd = "ALTER TABLE %s.%s ALTER COLUMN %s TYPE %s;" % (
            schema, table.lower(), column_name, column_type)
    else:
        cmd = "ALTER TABLE %s.%s ALTER COLUMN %s TYPE %s USING %s;" % (
            schema, table.lower(), column_name, column_type, expression)
    logger.warn("""
    [TABLE] ALTER TABLE %s, changing type of column '%s' to '%s'
    """ % (
        table.lower(), column_name, column_type))

    try:
        db.execute_cmd(cmd)
    except psycopg2.ProgrammingError as ex:
        logger.error(
            """
            [TABLE] ProgrammingError: %s when executing command %s.
            """ % (ex, cmd))
    except Exception as ex:
        logger.error('[TABLE] %s when executing command %s.' % (ex, cmd))


def remove_column(db, schema, table, column_name):
    """
    Remove a column from a table.
    Parameters
    ----------
    db: obj
    schema: str
    table: str
    column_name: str

    Example
    -------
    remove_column(pg, 'public', 'user', age')
    """
    cmd = "ALTER TABLE IF EXISTS %s.%s DROP COLUMN IF EXISTS %s;" % (
        schema, table.lower(), column_name)
    try:
        logger.warn("""
        [TABLE] Removing column '%s' from table '%s.%s' if exists.
        """ % (
            column_name, schema, table.lower()))
        db.execute_cmd(cmd)
    except Exception as ex:
        logger.error('[TABLE] %s when executing command %s.' % (ex, cmd))


def get_column_names_and_types(db, schema, table):
    """
    Get column names and column types of a specific table.
    Parameters
    ----------
    table_name: str
    Returns
    -------
    List of column names and corresponding types.
    """
    cmd = """
    SELECT column_name, data_type FROM information_schema.columns
    WHERE table_schema='%s' AND table_name = '%s';
    """ % (
        schema, table.lower())
    logger.info("[TABLE] Checking columns and types for table %s.%s" %
                (schema, table))
    try:
        rows = db.execute_cmd_with_fetch(cmd)
        return rows
    except Exception as ex:
        logger.error('[TABLE] %s when executing command %s.' % (ex, cmd))


def create_trigger_type_notification(db, schema, table, name, proc):
    cmd = """
    CREATE TRIGGER %s AFTER INSERT OR UPDATE OR DELETE ON %s.%s
    FOR EACH ROW EXECUTE PROCEDURE %s()
    """ % (name, schema, table, proc)
    try:
        logger.info("[TABLE] Creating trigger '%s'" % name)
        db.execute_cmd(cmd)
    except Exception as ex:
        logger.error("[TABLE] Creating trigger '%s' failed: %s" % (name, ex))


def drop_trigger_type_notification(db, schema, table, name, proc):
    cmd = "DROP TRIGGER IF EXISTS %s ON %s.%s CASCADE" % (name, schema, table)
    try:
        logger.info("[TABLE] Dropping trigger '%s'" % name)
        db.execute_cmd(cmd)
    except Exception as ex:
        logger.error("[TABLE] Dropping trigger '%s' failed: %s" % (name, ex))

def vacuum(db, schema, table):
    cmd = "VACUUM FULL ANALYZE %s.%s;" % (schema, table)
    try:
        logger.info("[TABLE] Vacuuming table '%s.%s'" % (schema, table))
        db.execute_cmd(cmd)    
    except Exception as ex:
        logger.error("[TABLE] Vacuuming table '%s.%s' failed: %s" % (schema, table))




from etl.transform import relation
from etl.extract import extractor
import unittest
from tests.meta import mock
from etl.extract import collection_map as cm
import copy

pg = mock.pg
mongo = mock.mongo
query = mock.query
rel_name_company = mock.rel_name_company
coll_name_company = mock.coll_name_company
coll_name_employee = mock.coll_name_employee
coll_conf_new = mock.coll_config_new
ex = extractor.Extractor(pg, mongo, mock.setup_pg,
                         mock.settings, mock.coll_config)
pg_cm_attrs = mock.pg_coll_map_attrs


def create_and_populate_company_pg():
    cursor = pg.conn.cursor()
    cursor.execute(query["table_drop_purr_cm"])
    # create table for CM in the database
    cm.create_table(pg, mock.coll_config)

    cursor.execute(query["table_drop_company"])
    cursor.execute(query["table_create_company"])

    cursor.execute(
        """insert into company(
            id, active, domains, signup_code
            ) values(
                '12345', 'true',
                '{"domain": ["pelotonland.com"]}',
                'xfLfdsFD3S')""")

    cursor.close()


def create_and_populate_company_mdb():
    # TODO
    mongo.drop_collection(coll_name_company)
    print("creates collection company")
    docs = mock.data_mdb_company
    mongo[coll_name_company].insert_many(docs)


def create_and_populate_employee_mdb():
    mongo.drop_collection(coll_name_employee)
    print("creates collection employee")
    docs = mock.data_mdb_employee
    mongo[coll_name_employee].insert_many(docs)


class TestExtractor(unittest.TestCase):

    def test_init_db(self):
        # create test databases in Mongo and Postgres
        # Postgres
        cursor = pg.conn.cursor()
        cursor.execute(query["db_exists"])
        db_exists = cursor.fetchone()
        assert db_exists[0] is True
        cursor.close()

    def test_update_table_def_change_type(self):
        # Check if column was changed
        # update domain type: jsonb to text
        res_expected = [('active', 'boolean'), ('domains', 'text'),
                        ('id', 'text'), ('signup_code', 'text')]

        create_and_populate_company_pg()
        cursor = pg.conn.cursor()

        # update type
        ex.update_table_def(mock.coll_config_db, mock.coll_config_db_new)

        # get columns and types from information schema
        cursor.execute(query["table_check_company_columns"])
        res_db = cursor.fetchall()

        print(res_db)
        print(res_expected)
        cursor.close()
        assert res_db == res_expected

    def test_update_table_def_add_column(self):
        res_expected = [
            ('active', 'boolean'), ('domains', 'jsonb'),
            ('id', 'text'), ('signup_code', 'text'),
            ('updated_at', 'timestamp without time zone')
        ]

        create_and_populate_company_pg()
        cursor = pg.conn.cursor()

        # add type
        conf = mock.coll_config_db
        conf_new = copy.deepcopy(conf)

        conf_new[0][3].append({
            ':source': 'updatedAt',
            ':type': 'TIMESTAMP',
            'updated_at': None
        })

        ex.update_table_def(conf, conf_new)

        # get columns and types from information schema
        cursor.execute(query["table_check_company_columns"])
        res_db = cursor.fetchall()

        print(res_db)
        print(res_expected)

        cursor.close()
        assert res_db == res_expected

    def test_update_table_def_remove_column(self):
        # Removing last column (signup code)
        res_expected = [('active', 'boolean'), ('domains', 'jsonb'),
                        ('id', 'text')]

        create_and_populate_company_pg()
        cursor = pg.conn.cursor()

        # add type
        conf = copy.deepcopy(mock.coll_config_db)
        conf_new = copy.deepcopy(conf)

        conf_new[0][3].pop()

        ex.update_table_def(conf, conf_new)

        # get columns and types from information schema
        cursor.execute(query["table_check_company_columns"])
        res_db = cursor.fetchall()

        print(res_db)
        print(res_expected)

        cursor.close()
        assert res_db == res_expected

    def test_table_track(self):
        # TODO: see if the table is transfered to the PG database
        create_and_populate_company_mdb()
        create_and_populate_employee_mdb()

        cursor = pg.conn.cursor()
        cursor.execute(query["table_drop_company"])
        cursor.execute(query["table_drop_employee"])
        cursor.execute(query["table_drop_purr_cm"])

        coll_config = copy.deepcopy(mock.coll_config)
        # create table for CM in the database
        cm.create_table(pg, mock.coll_config)
        ex = extractor.Extractor(
            pg, mongo, mock.setup_pg, mock.settings, coll_config)

        # # changes extractor's collection definition for every collection
        # # and transfers

        coll_map_old = mock.coll_config_db
        coll_map_new = mock.coll_config_db_company_employee
        ex.table_track(coll_map_old, coll_map_new)

        mocked = mock.coll_config_company_employee
        if (len(ex.coll_def) != len(mocked)):
            assert False

        for k, v in ex.coll_def.items():
            if v != mocked[k]:
                print(k)
                print(v[":columns"])
                print(mocked[k][":columns"])
                if v[":columns"] != mocked[k][":columns"]:
                    assert False

        cursor.execute(query["table_drop_purr_cm"])
        cursor.close()
        del ex

        assert True

    def test_table_untrack(self):
        # TODO:
        # - check if the table is left in the PG database
        # - try to insert new data to mongodb and check
        # if its left out from the data transfer
        # this one should remove employee from the collection map
        create_and_populate_company_mdb()
        create_and_populate_employee_mdb()

        cursor = pg.conn.cursor()
        cursor.execute(query["table_drop_company"])
        cursor.execute(query["table_drop_employee"])
        cursor.execute(query["table_drop_purr_cm"])

        # create table for CM in the database
        cm.create_table(pg, mock.coll_config)
        coll_config = copy.deepcopy(mock.coll_config_company_employee)

        ex = extractor.Extractor(
            pg,
            mongo,
            mock.setup_pg,
            mock.settings,
            coll_config
        )

        # # changes extractor's collection definition for every collection
        # # and transfers

        coll_map_old = copy.deepcopy(mock.coll_config_db_company_employee)
        coll_map_new = copy.deepcopy(mock.coll_config_db)
        ex.table_untrack(coll_map_old, coll_map_new)

        mocked = copy.deepcopy(mock.coll_config)
        if (len(ex.coll_def) != len(mocked)):
            print("NEW", ex.coll_def)
            print("OlD", mocked)
            assert False

        for k, v in ex.coll_def.items():
            if v != mocked[k]:
                print(k)
                print(v[":columns"])
                print(mocked[k][":columns"])
                if v[":columns"] != mocked[k][":columns"]:
                    assert False

        cursor.close()
        del ex
        assert True

    def test_update_coll_map_unchanged(self):

        # collection map is not changed
        # extractor.coll_map_cur stays the same

        cursor = pg.conn.cursor()

        cursor.execute(query["table_drop_purr_cm"])

        # create table for CM in the database
        cm.create_table(pg, mock.coll_config)

        coll_config = copy.deepcopy(mock.coll_config)
        ex = extractor.Extractor(
            pg,
            mongo,
            mock.setup_pg,
            mock.settings,
            coll_config
        )
        ex.update_coll_map()
        print("OLD", ex.coll_map_cur)
        print("NEW", mock.coll_config_db)
        cursor.close()

        res = (ex.coll_map_cur == mock.coll_config_db)
        del ex
        assert res

    def test_update_coll_map_changed(self):

        # collection map is changed and
        # extractor.coll_map_cur needs to be updated

        # purr_collection_map needs old values
        cursor = pg.conn.cursor()
        cursor.execute(query["table_drop_purr_cm"])

        # create table for CM in the database
        cm.create_table(pg, mock.coll_config)

        coll_config = copy.deepcopy(mock.coll_config_company_employee)

        ex = extractor.Extractor(
            pg,
            mongo,
            mock.setup_pg,
            mock.settings,
            coll_config
        )

        # this pulls the map from the db (mock.coll_config)
        ex.update_coll_map()
        print("NEW", ex.coll_map_cur)
        print("MOCK", mock.coll_config_db_company_employee)
        # case 2: collection map is changed and
        # extractor.coll_map_cur needs to be updated
        cursor.close()

        res = (ex.coll_map_cur == mock.coll_config_db)
        del ex
        assert res

    def test_transfer(self):
        # drop/truncate table
        # create schema
        # transfers collections
        cursor = pg.conn.cursor()

        # reset CM in the database
        cursor.execute(query["table_drop_purr_cm"])
        cursor.execute(query["table_drop_company"])
        cursor.execute(query["table_drop_employee"])
        create_and_populate_company_mdb()
        cm.create_table(pg, mock.coll_config)

        # collection which will be transferred
        collection_names = mock.coll_names

        coll_config = copy.deepcopy(mock.coll_config_company_employee)
        ex = extractor.Extractor(
            pg,
            mongo,
            mock.setup_pg,
            mock.settings,
            coll_config
        )
        ex.transfer(collection_names)
        for i in range(len(mock.rel_names)):
            relation = mock.rel_names[i]
            cursor.execute("SELECT count(*) FROM %s" % (relation))
            cnt_pg = cursor.fetchone()
            cnt_mongo = mongo[collection_names[i]].count()
            if cnt_mongo != cnt_pg[0]:
                print("Postgres:", cnt_pg[0])
                print("MongoDB:", cnt_mongo)
                assert False

        cursor.close()
        del ex

        assert True

    def test_transfer_coll(self):
        # TODO: check

        cursor = pg.conn.cursor()

        # reset CM in the database
        cursor.execute(query["table_drop_purr_cm"])
        cursor.execute(query["table_drop_company"])
        cursor.execute(query["table_drop_employee"])
        create_and_populate_company_mdb()
        cm.create_table(pg, mock.coll_config)

        # collection which will be transferred
        collection = mock.coll_names[0]
        relation = mock.rel_names[0]

        coll_config = copy.deepcopy(mock.coll_config_company_employee)
        ex = extractor.Extractor(
            pg,
            mongo,
            mock.setup_pg,
            mock.settings,
            coll_config
        )
        ex.transfer_coll(collection)
        cursor.execute("SELECT count(*) FROM %s" % (relation))
        cnt_pg = cursor.fetchone()
        cnt_mongo = mongo[collection].count()
        print("Postgres:", cnt_pg[0])
        print("MongoDB:", cnt_mongo)

        cursor.close()
        del ex
        assert cnt_mongo == cnt_pg[0]

    def test_insert_multiple(self):
        # no unset
        # TODO: test when there is value in unset
        cursor = pg.conn.cursor()

        # reset CM in the database
        cursor.execute(query["table_drop_purr_cm"])
        cursor.execute(query["table_drop_company"])

        create_and_populate_company_mdb()
        cm.create_table(pg, mock.coll_config)

        coll = mock.coll_names[0]
        rel = mock.rel_names[0]
        docs = []

        for doc in mongo[coll].find():
            docs.append(doc)
        coll_config = copy.deepcopy(mock.coll_config_company_employee)
        ex = extractor.Extractor(
            pg,
            mongo,
            mock.setup_pg,
            mock.settings,
            coll_config
        )
        schema = mock.setup_pg["schema_name"]

        r = relation.Relation(
            pg,
            schema,
            rel,
            True
        )
        attrs = mock.attrs_company
        types = mock.types_company
        r.create_with_columns(attrs, types)

        ex.insert_multiple(docs, r, coll)

        print("COLLECTION", coll)
        print("DOCUMENTS", docs)

        cmd = "SELECT %s FROM %s order by id" % (", ".join(attrs[1:]), rel)
        cursor.execute(cmd)
        mocked = mock.data_pg_company_no_id
        res = cursor.fetchall()
        print("MOCKED")
        print(mocked)
        print("RESULT")
        print(res)

        cursor.close()
        del r
        del ex
        assert mocked == res

    def test_update_multiple(self):
        # no unset
        # TODO: test when there is value in unset
        cursor = pg.conn.cursor()

        # reset CM in the database
        cursor.execute(query["table_drop_purr_cm"])
        cursor.execute(query["table_drop_company"])

        create_and_populate_company_mdb()
        cm.create_table(pg, mock.coll_config)

        coll = mock.coll_names[0]
        rel = mock.rel_names[0]
        docs = []
        mock_updated = mock.data_mdb_company_updated
        for doc in mongo[coll].find():
            docs.append({
                "_id": doc["_id"],
                "active": mock_updated["active"],
                "signupCode": mock_updated["signupCode"],
                "domains": mock_updated["domains"]
            })
        coll_config = copy.deepcopy(mock.coll_config_company_employee)
        ex = extractor.Extractor(
            pg,
            mongo,
            mock.setup_pg,
            mock.settings,
            coll_config
        )
        schema = mock.setup_pg["schema_name"]

        r = relation.Relation(
            pg,
            schema,
            rel,
            True
        )
        attrs = mock.attrs_company
        types = mock.types_company
        r.create_with_columns(attrs, types)

        ex.update_multiple(docs, r, coll)

        cmd = "SELECT %s FROM %s order by id" % (", ".join(attrs[1:]), rel)
        cursor.execute(cmd)
        mocked = mock.data_pg_company_updated_no_id
        res = cursor.fetchall()
        print("MOCKED")
        print(mocked)
        print("RESULT")
        print(res)

        cursor.close()
        del r
        del ex
        assert mocked == res

    def test_prepare_attr_details_with_extra_props(self):
        coll_config = copy.deepcopy(mock.coll_config_company_employee)
        ex = extractor.Extractor(
            pg,
            mongo,
            mock.setup_pg,
            mock.settings,
            coll_config
        )
        type_extra_props_pg = 'jsonb'

        ex.include_extra_props = True
        attrs = copy.deepcopy(mock.attrs_company)
        fields = copy.deepcopy(mock.fields_company)
        types = []
        for item in mock.types_company:
            types.append(item.lower())

        attribute_details = ex.prepare_attr_details(
            attrs, fields, types, type_extra_props_pg)

        print("RESULT")
        print(attribute_details)
        print("MOCKED")
        print(mock.attr_details)
        del ex
        assert mock.attr_details == attribute_details

    def test_prepare_attr_details_without_extra_props(self):
        coll_config = copy.deepcopy(mock.coll_config_company_employee)
        ex = extractor.Extractor(
            pg,
            mongo,
            mock.setup_pg,
            mock.settings,
            coll_config
        )

        ex.include_extra_props = False
        attrs = copy.deepcopy(mock.attrs_company)
        fields = copy.deepcopy(mock.fields_company)
        types = []
        for item in mock.types_company:
            types.append(item.lower())

        attribute_details = ex.prepare_attr_details(
            attrs, fields, types)

        print("RESULT")
        print(attribute_details)
        print("MOCKED")
        print(mock.attr_details)
        del ex
        mocked = copy.deepcopy(mock.attr_details)
        del mocked["extraProps"]

        assert mocked == attribute_details

    # def test_adjust_columns(self):
    #     assert False

    # def test_handle_failed_type_update(self):
    #     # self.attrs_details changed
    #     assert False

    # def test_add_extra_props(self):
    #     assert False



from setuptools import setup, find_packages
from os import path
this_directory = path.abspath(path.dirname(__file__))

with open('README.md', encoding='utf-8') as f:
    long_description = f.read()

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

version = '1.0.3'

setup(name='purrito',
      version=version,
      author='Anett Balázsics',
      email='balazsicsanett@gmail.com',
      description="ETL tool for transfering data from MongoDB to PostgreSQL",
      url='https://boosterfuels.github.io/purr',
      packages=find_packages(),
      classifiers=[
          "Programming Language :: Python :: 3",
          "License :: OSI Approved :: MIT License",
          "Operating System :: OS Independent",
          "Topic :: Database"
      ],
      install_requires=requirements,
      entry_points={
          'console_scripts': [
              'purrito = etl.__main__:main'
          ]
      },
      long_description=long_description,
      long_description_content_type='text/markdown'
      )



from threading import Thread
from etl.monitor import logger
import time

class NotificationThread(Thread):
    conn = None
    new = False

    def __init__(self, conn):
        Thread.__init__(self)
        self.conn = conn
        self.terminate = False

    def run(self):
        self.listen()

    def stop(self):
        self.terminate = True

    def listen(self):
        """
        Listens for changes from the channel.
        Parameters
        ----------
        db :    pgection
        Returns
        -------
        -
        Example
        -------
        listen(pg)
        """

        channel = 'purr'
        cmd = 'LISTEN %s;' % channel
        db = self.conn.conn
        db.cursor().execute(cmd)
        running = True
        try:
            while running:
                time.sleep(1)
                if self.terminate is True:
                    break
                db.commit()
                db.poll()
                db.commit()
                while db.notifies:
                    notify = db.notifies.pop()
                    logger.info("""[NOTIFICATION]
                     New notification: pid=%s channel=%s payload=%s new=%s
                    """ % (
                        notify.pid, notify.channel, notify.payload, self.new))
                    self.new = True
        except Exception as ex:
            logger.error("[NOTIFICATION] Details: %s" % ex)
